# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Color selector functionality. E.g. for use in the plugin config dialog
"""
import logging
from argos.qt import QtCore, QtGui, QtWidgets, Qt
from argos.utils.defs import InvalidInputError

logger = logging.getLogger(__name__)


class ShortCutEditor(QtWidgets.QLineEdit):

    def __init__(self, parent=None):
        super(ShortCutEditor, self).__init__(parent=parent)


    def keyPressEvent(self, event):

        key = event.key()

        if (Qt.Key_A <= key <= Qt.Key_Z or Qt.Key_0 <= key <= Qt.Key_9 or
            Qt.Key_F1 <= key <= Qt.Key_F35):

            seq = QtGui.QKeySequence(event.modifiers()|event.key()).toString()

            logger.debug("Key press event: {}: {}".format(event.key(), seq))
            self.setText(seq)
        else:
            super(ShortCutEditor, self).keyPressEvent(event)
        return



"""This module provides functions to parse JSON-with-comments.

Standard JSON is a human-readable data interchange format that is popular due to its simplicity
and wide support across programming languages. Unfortunately, it does not support comments,
making it less suitable for some applications like configuration files.

This module implements several functions that add support for comments to JSON, resulting in a
JSON dialect that we refer to as "JSON-with-comments"; elsewhere, this format has been referred
to as "jsonc".

Two types of comments are supported, mirroring the two types of comments available in the
JavaScript language from which JSON originates:

* Line comments start with "//" and continue to the next newline character;
* Block comments start with "/*" and end with the next "*/".

The core functionality of this module is provided by the erase_json_comments() function. It
takes an input string argument, replaces comments with whitespace, and returns an output string
that is 'pure' JSON. Inside comments, carriage returns and newlines are passed through as-is;
all other characters are replaced by spaces. This preserves the structure of the original input,
allowing a subsequently run JSON parser to report any issues in the output with line numbers and
column offsets that correspond to the original JSON-with-comments input.

Two convenience functions are provided that run erase_json_comments() on input, then pass the
resulting 'purified' JSON-string to the JSON parser that is provided by the 'json' module in the
Python standard library. For most programs, one of these two functions will provide the easiest
way to use JSON-with-comments:

* parse_json_with_comments_string() parses JSON-with comments from a string;
* parse_json_with_comments_file() parses JSON-with comments from a file.
"""

import json
from enum import Enum


class JSONWithCommentsError(ValueError):
    """An error occurred while scanning JSON-with-comments."""


class FsmState(Enum):
    """Finite State Machine states for scanning JSON-with-comments."""
    DEFAULT            = 1  # default state ; not scanning a string or comment.
    COMMENT_INTRO      = 2  # accepted '/'  ; the start of a line or block comment.
    LINE_COMMENT       = 3  # accepted '//' ; scanning a line comment.
    BLOCK_COMMENT      = 4  # accepted '/*' ; scanning a block comment.
    BLOCK_COMMENT_STAR = 5  # accepted '/*' ; scanning a block comment, and just scanned a '*'.
    STRING             = 6  # accepted '"'  ; scanning a string.
    STRING_BACKSLASH   = 7  # accepted '"'  ; scanning a string, and just scanned a backslash.


class FsmAction(Enum):
    """Finite State Machine actions for scanning JSON-with-comments."""
    EMIT_NOTHING             = 1  # emit nothing
    EMIT_CURRENT             = 2  # emit the current character
    EMIT_FSLASH_THEN_CURRENT = 3  # emit forward slash followed by current character
    EMIT_ONE_SPACE           = 4  # emit one space character
    EMIT_TWO_SPACES          = 5  # emit two space characters


class CharacterClass(Enum):
    """Character classes to be distinguished while scanning JSON-with-comments."""
    FSLASH = 1  # forward slash
    BSLASH = 2  # back slash
    QUOTE  = 3  # string quote
    CR     = 4  # carriage return
    NL     = 5  # newline
    STAR   = 6  # star (asterisk)
    OTHER  = 7  # any other character


# Define the transition table of (state, character_class) -> (action, next_state).
#
# Note that not all possible combinations of (state, character_class) are explicitly specified.
# If a (state, character_class) combination is encountered while scanning that is not present
# in the transition table, the key (state, CharacterClass.OTHER) key will be used instead.
#
# This makes the transition table both smaller (only 22 out of 49 entries need to be specified)
# and easier to understand.

_fsm_definition = {

    ( FsmState.DEFAULT            , CharacterClass.FSLASH ) : ( FsmAction.EMIT_NOTHING             , FsmState.COMMENT_INTRO      ),
    ( FsmState.DEFAULT            , CharacterClass.QUOTE  ) : ( FsmAction.EMIT_CURRENT             , FsmState.STRING             ),
    ( FsmState.DEFAULT            , CharacterClass.OTHER  ) : ( FsmAction.EMIT_CURRENT             , FsmState.DEFAULT            ),

    ( FsmState.COMMENT_INTRO      , CharacterClass.FSLASH ) : ( FsmAction.EMIT_TWO_SPACES          , FsmState.LINE_COMMENT       ),
    ( FsmState.COMMENT_INTRO      , CharacterClass.STAR   ) : ( FsmAction.EMIT_TWO_SPACES          , FsmState.BLOCK_COMMENT      ),
    ( FsmState.COMMENT_INTRO      , CharacterClass.OTHER  ) : ( FsmAction.EMIT_FSLASH_THEN_CURRENT , FsmState.DEFAULT            ),

    ( FsmState.LINE_COMMENT       , CharacterClass.CR     ) : ( FsmAction.EMIT_CURRENT             , FsmState.LINE_COMMENT       ),
    ( FsmState.LINE_COMMENT       , CharacterClass.NL     ) : ( FsmAction.EMIT_CURRENT             , FsmState.DEFAULT            ),
    ( FsmState.LINE_COMMENT       , CharacterClass.OTHER  ) : ( FsmAction.EMIT_ONE_SPACE           , FsmState.LINE_COMMENT       ),

    ( FsmState.BLOCK_COMMENT      , CharacterClass.CR     ) : ( FsmAction.EMIT_CURRENT             , FsmState.BLOCK_COMMENT      ),
    ( FsmState.BLOCK_COMMENT      , CharacterClass.NL     ) : ( FsmAction.EMIT_CURRENT             , FsmState.BLOCK_COMMENT      ),
    ( FsmState.BLOCK_COMMENT      , CharacterClass.STAR   ) : ( FsmAction.EMIT_ONE_SPACE           , FsmState.BLOCK_COMMENT_STAR ),
    ( FsmState.BLOCK_COMMENT      , CharacterClass.OTHER  ) : ( FsmAction.EMIT_ONE_SPACE           , FsmState.BLOCK_COMMENT      ),

    ( FsmState.BLOCK_COMMENT_STAR , CharacterClass.FSLASH ) : ( FsmAction.EMIT_ONE_SPACE           , FsmState.DEFAULT            ),
    ( FsmState.BLOCK_COMMENT_STAR , CharacterClass.CR     ) : ( FsmAction.EMIT_CURRENT             , FsmState.BLOCK_COMMENT      ),
    ( FsmState.BLOCK_COMMENT_STAR , CharacterClass.NL     ) : ( FsmAction.EMIT_CURRENT             , FsmState.BLOCK_COMMENT      ),
    ( FsmState.BLOCK_COMMENT_STAR , CharacterClass.STAR   ) : ( FsmAction.EMIT_ONE_SPACE           , FsmState.BLOCK_COMMENT_STAR ),
    ( FsmState.BLOCK_COMMENT_STAR , CharacterClass.OTHER  ) : ( FsmAction.EMIT_ONE_SPACE           , FsmState.BLOCK_COMMENT      ),

    ( FsmState.STRING             , CharacterClass.BSLASH ) : ( FsmAction.EMIT_CURRENT             , FsmState.STRING_BACKSLASH   ),
    ( FsmState.STRING             , CharacterClass.QUOTE  ) : ( FsmAction.EMIT_CURRENT             , FsmState.DEFAULT            ),
    ( FsmState.STRING             , CharacterClass.OTHER  ) : ( FsmAction.EMIT_CURRENT             , FsmState.STRING             ),

    ( FsmState.STRING_BACKSLASH   , CharacterClass.OTHER  ) : ( FsmAction.EMIT_CURRENT             , FsmState.STRING             )
}

# Characters are classified according to this table.
# All characters that are not explicitly specified belong to CharacterClass.OTHER.

_character_classifications = {
    '/'  : CharacterClass.FSLASH,
    '\\' : CharacterClass.BSLASH,
    '\"' : CharacterClass.QUOTE,
    '\r' : CharacterClass.CR,
    '\n' : CharacterClass.NL,
    '*'  : CharacterClass.STAR
}


def erase_json_comments(input_string: str) -> str:
    """Erase the comments in a JSON-with-comments string and return valid JSON.

    All characters in a comment are replaced by spaces, except that carriage return and newline
    characters are passed through.

    This ensures that the structure of the output is identical to the structure of the input,
    allowing a subsequently run JSON parser to report any issues with meaningful line numbers
    and column offsets.
    """

    output = []

    state = FsmState.DEFAULT

    for current_character in input_string:

        character_class = _character_classifications.get(current_character, CharacterClass.OTHER)

        # If a (state, character_class) tuple is not explicitly handled in the FSM definition,
        # the behavior of CharacterClass.OTHER will be used.

        if (state, character_class) not in _fsm_definition:
            character_class = CharacterClass.OTHER

        # Look up action and next state.

        (action, next_state) = _fsm_definition[(state, character_class)]

        # Perform the specified action.

        if action in (FsmAction.EMIT_CURRENT, FsmAction.EMIT_FSLASH_THEN_CURRENT):
            if action == FsmAction.EMIT_FSLASH_THEN_CURRENT:
                output.append("/")
            output.append(current_character)
        elif action in (FsmAction.EMIT_ONE_SPACE, FsmAction.EMIT_TWO_SPACES):
            if action == FsmAction.EMIT_TWO_SPACES:
                output.append(" ")
            output.append(" ")

        # Proceed to the next state.

        state = next_state

    # We're at the end of the character scan loop.

    # The usual end state for a successful scan will be DEFAULT. We also permit LINE_COMMENT,
    # meaning that line comments that are not terminated by a newline are acceptable.
    #
    # If we find ourselves in one of the five other states at the end of the scan, something is
    # wrong.
    #
    # If we're in either the STRING or STRING_BACKSLASH state, the input ended while scanning a
    # string, and this will also be true in the output we produce. A JSON parser that consumes
    # our output will detect this error, so no action is needed here to handle this case.
    #
    # The three remaining erroneous states do require action from our side:

    if state == FsmState.COMMENT_INTRO:
        # If we're in the COMMENT_INTRO state, the input ended in a forward slash. This forward
        # slash was not yet emitted when we entered the COMMENT_INTRO state, so we emit it now.
        # Since no valid JSON file can end with a forward slash, this slash will be caught when
        # a JSON parser processes our output, causing a parsing error.
        output.append("/")
    elif state in (FsmState.BLOCK_COMMENT, FsmState.BLOCK_COMMENT_STAR):
        # If we're in either the BLOCK_COMMENT or BLOCK_COMMENT_STAR state, the input ended while
        # scanning an unterminated block comment. This issue would not be noticed by a JSON parser
        # that processes our output, as our output will just end with the spaces that replaced the
        # partial block comment. So for both these end states, we will raise an exception here.
        raise JSONWithCommentsError("Unterminated block comment.")

    # Concatenate the output characters and return the result.

    output_string = "".join(output)

    return output_string


def parse_json_with_comments_string(json_with_comments: str, **kw):
    """Parse a JSON-with-comments string by erasing comments and parsing the result as JSON.

    Any keyword arguments are passed on to the json.loads() function that is used to parse
    the input string from which the comments have been erased.
    """
    json_without_comments = erase_json_comments(json_with_comments)
    try:
        return json.loads(json_without_comments, **kw)
    except json.JSONDecodeError as json_exception:
        # Wrap the JSONDecodeError in a JSONWithCommentsError exception.
        raise JSONWithCommentsError("Invalid JSON after erasing comments.") from json_exception


def parse_json_with_comments_file(filename: str, **kw):
    """Parse a JSON-with-comments file by erasing comments and parsing the result as JSON.

    Any keyword arguments are passed on to the json.loads() function that is used to parse
    the input string as read from the file from which the comments have been erased.
    """
    with open(filename, "r") as input_file:
        json_with_comments = input_file.read()
    return parse_json_with_comments_string(json_with_comments, **kw)



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Constants related to the layout and other widget properties.
"""
import sys
from argos.qt import QtCore, QtGui, QtWidgets

# Use large numpy line length in the Quicklook panel
NUMPY_LINE_WIDTH = sys.maxsize

#TREE_ROW_HEIGHT = 20 # pixels
TREE_CELL_SIZE_HINT = QtCore.QSize(100, 20)
TREE_ICON_SIZE = QtCore.QSize(16, 16)

#COLLECTOR_TREE_CELL_SIZE_HINT = QtCore.QSize(100, 24) not used (yet?)
COLLECTOR_TREE_ICON_SIZE = QtCore.QSize(20, 20)

# Initial dock widths of the main window
LEFT_DOCK_WIDTH = 440  # needs room for scroll bars
RIGHT_DOCK_WIDTH = 320
TOP_DOCK_HEIGHT = 75

COL_NODE_NAME_WIDTH = 180
COL_KIND_WIDTH = 50
COL_ELEM_TYPE_WIDTH = 80
COL_SUMMARY_WIDTH = 110 # will be stretched


# Spacing and margin in central widgets in pixels
CENTRAL_SPACING = 0
CENTRAL_MARGIN  = 0

# Spacing and margin in dock widgets in pixels. Is now set in style sheet margin.

DOCK_SPACING = 0
DOCK_MARGIN  = 0

if sys.platform == 'linux':
    MONO_FONT = 'Monospace'
    FONT_SIZE = 10
elif sys.platform == 'win32' or sys.platform == 'cygwin':
    MONO_FONT = 'Courier'
    FONT_SIZE = 10
elif sys.platform == 'darwin':
    MONO_FONT = 'Courier'
    FONT_SIZE = 13
else:
    MONO_FONT = 'Courier'
    FONT_SIZE = 13

COLOR_ERROR = '#FF0000' # red

QCOLOR_REGULAR = QtGui.QColor('black')
QCOLOR_NOT_IMPORTED = QtGui.QColor('grey')
QCOLOR_ERROR = QtGui.QColor(COLOR_ERROR)




# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Configuration tree. Can be used to manipulate the a ConfigurationModel.

"""
from __future__ import print_function

import enum
import logging
import os.path

from argos.config.abstractcti import ResetMode
from argos.config.configitemdelegate import ConfigItemDelegate
from argos.config.configtreemodel import ConfigTreeModel
from argos.info import DEBUGGING, icons_directory
from argos.qt import Qt, QtCore, QtGui, QtWidgets, QtSlot
from argos.widgets.argostreeview import ArgosTreeView
from argos.widgets.constants import RIGHT_DOCK_WIDTH, DOCK_SPACING, DOCK_MARGIN
from argos.widgets.misc import BasePanel
from argos.utils.cls import check_class

logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901



class ConfigWidget(BasePanel):
    """ Shows the configuration. At the moment only the confg tree view.
    """

    def __init__(self, configTreeModel, parent=None):
        """ Constructor.
            :param parent:
        """
        super(ConfigWidget, self).__init__(parent=parent)

        # Actions that change the reset mode of the reset button
        self.modeActionGroup = QtWidgets.QActionGroup(self)
        self.modeActionGroup.setExclusive(True)

        self.modeAllAction = QtWidgets.QAction("Reset All", self.modeActionGroup)
        self.modeAllAction.setToolTip("Changes button reset mode to reset all settings")
        self.modeAllAction.setCheckable(True)
        self.modeAllAction.triggered.connect(lambda : self.setResetMode(ResetMode.All))


        self.modeRangeAction = QtWidgets.QAction("Reset Ranges", self.modeActionGroup)
        self.modeRangeAction.setToolTip("Changes button reset mode to reset axes")
        self.modeRangeAction.setCheckable(True)
        self.modeRangeAction.triggered.connect(lambda : self.setResetMode(ResetMode.Ranges))

        # Sanity check that actions have been added to action group
        assert self.modeActionGroup.actions(), "Sanity check. resetActionGroup is empty"

        # Actions that actually reset the settings

        self.resetAllAction = QtWidgets.QAction("Reset All", self)
        self.resetAllAction.setToolTip("Resets all settings.")
        self.resetAllAction.setIcon(QtGui.QIcon(os.path.join(icons_directory(), 'reset-l.svg')))
        self.resetAllAction.setShortcut("Ctrl+=")

        self.resetRangesAction = QtWidgets.QAction("Reset Ranges", self)
        self.resetRangesAction.setToolTip(
            "Resets range of all plots, color scales, table column/row sizes etc.")
        self.resetRangesAction.setIcon(QtGui.QIcon(os.path.join(icons_directory(), 'reset-l.svg')))
        self.resetRangesAction.setShortcut("Ctrl+0")

        self.resetButtonMenu = QtWidgets.QMenu()
        self.resetButtonMenu.addAction(self.resetAllAction)
        self.resetButtonMenu.addAction(self.resetRangesAction)
        self.resetButtonMenu.addSection("Default")
        self.resetButtonMenu.addAction(self.modeAllAction)
        self.resetButtonMenu.addAction(self.modeRangeAction)

        # Widgets

        self.mainLayout = QtWidgets.QVBoxLayout(self)
        self.mainLayout.setSpacing(5)
        self.mainLayout.setContentsMargins(DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN)

        self.configTreeView = ConfigTreeView(configTreeModel, parent=self)
        self.mainLayout.addWidget(self.configTreeView)

        self.buttonLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.addLayout(self.buttonLayout)

        self.autoCheckBox = QtWidgets.QCheckBox("Auto")
        self.autoCheckBox.setToolTip("Auto reset when a new item or axis is selected.")
        self.autoCheckBox.setChecked(True)

        self.resetButton = QtWidgets.QToolButton()
        self.resetButton.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.resetButton.setDefaultAction(self.resetButtonMenu.defaultAction())
        self.resetButton.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
        self.resetButton.setMenu(self.resetButtonMenu)

        # Set font size to the same as used for push buttons
        dummyButton = QtWidgets.QPushButton("dummy")
        fontSize = dummyButton.font().pointSize()
        del dummyButton

        logger.debug("Setting QToolButtons font size to: {} point".format(fontSize))
        font = self.resetButton.font()
        font.setPointSizeF(fontSize)
        self.resetButton.setFont(font)

        self.buttonLayout.addStretch()
        self.buttonLayout.addWidget(self.autoCheckBox)
        self.buttonLayout.addWidget(self.resetButton)
        self.buttonLayout.addStretch()

        self.autoCheckBox.stateChanged.connect(self.setAutoReset)
        self.resetRangesAction.triggered.connect(self.configTreeView.resetAllRanges)
        self.resetAllAction.triggered.connect(self.configTreeView.resetAllSettings)

        self.setResetMode(self.configTreeView.resetMode)


    def setAutoReset(self, value):
        """ Sets config tree widget auto reset
        """
        self.configTreeView.autoReset = value


    def setResetMode(self, resetMode):
        """ Sets the reset mode of the reset button reset

            :param resetMode: 'Ranges' all 'All'
        """
        check_class(resetMode, ResetMode)
        if resetMode == ResetMode.All:
            self.resetButton.setDefaultAction(self.resetAllAction)
            self.modeAllAction.setChecked(True)
        elif resetMode == ResetMode.Ranges:
            self.resetButton.setDefaultAction(self.resetRangesAction)
            self.modeRangeAction.setChecked(True)
        else:
            raise ValueError("Unexpected resetMode: {}".format(resetMode))

        self.configTreeView.resetMode = resetMode


    def marshall(self):
        """ Returns a dictionary to save in the persistent settings
        """
        cfg = dict(
            autoRange = self.autoCheckBox.isChecked(),
            resetMode = self.configTreeView.resetMode.value,
        )
        return cfg


    def unmarshall(self, cfg):
        """ Initializes itself from a config dict form the persistent settings.
        """
        if 'autoRange' in cfg:
            self.autoCheckBox.setChecked(cfg['autoRange'])

        if 'resetMode' in cfg:
            self.setResetMode(ResetMode(cfg['resetMode']))



class ConfigTreeView(ArgosTreeView):
    """ Tree widget for manipulating a tree of configuration options.
    """
    def __init__(self, configTreeModel, parent=None):
        """ Constructor
        """
        super(ConfigTreeView, self).__init__(treeModel=configTreeModel, parent=parent)

        self._configTreeModel = configTreeModel

        self.expanded.connect(configTreeModel.expand)
        self.collapsed.connect(configTreeModel.collapse)
        #configTreeModel.update.connect(self.update) # not necessary
        #self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)

        treeHeader = self.header()
        treeHeader.resizeSection(ConfigTreeModel.COL_NODE_NAME, round(RIGHT_DOCK_WIDTH * 0.5))
        treeHeader.resizeSection(ConfigTreeModel.COL_VALUE, round(RIGHT_DOCK_WIDTH * 0.5))

        headerNames = self.model().horizontalHeaders
        enabled = dict((name, True) for name in headerNames)
        enabled[headerNames[ConfigTreeModel.COL_NODE_NAME]] = False # Name cannot be unchecked
        enabled[headerNames[ConfigTreeModel.COL_VALUE]] = False # Value cannot be unchecked
        checked = dict((name, False) for name in headerNames)
        checked[headerNames[ConfigTreeModel.COL_NODE_NAME]] = True # Checked by default
        checked[headerNames[ConfigTreeModel.COL_VALUE]] = True # Checked by default
        self.addHeaderContextMenu(checked=checked, enabled=enabled, checkable={})

        self.setRootIsDecorated(True)
        self.setUniformRowHeights(True)
        self.setItemDelegate(ConfigItemDelegate())
        self.setEditTriggers(QtWidgets.QAbstractItemView.AllEditTriggers)

        #self.setEditTriggers(QtWidgets.QAbstractItemView.DoubleClicked |
        #                     QtWidgets.QAbstractItemView.EditKeyPressed |
        #                     QtWidgets.QAbstractItemView.AnyKeyPressed |
        #                     QtWidgets.QAbstractItemView.SelectedClicked)


    def sizeHint(self):
        """ The recommended size for the widget."""
        return QtCore.QSize(RIGHT_DOCK_WIDTH, 500)


    @QtSlot(QtWidgets.QWidget, QtWidgets.QAbstractItemDelegate.EndEditHint)
    def closeEditor(self, editor, hint):
        """ Finalizes, closes and releases the given editor.
        """
        # It would be nicer if this method was part of ConfigItemDelegate since createEditor also
        # lives there. However, QAbstractItemView.closeEditor is sometimes called directly,
        # without the QAbstractItemDelegate.closeEditor signal begin emitted, e.g when the
        # currentItem changes. Therefore we cannot connect the QAbstractItemDelegate.closeEditor
        # signal to a slot in the ConfigItemDelegate.
        configItemDelegate = self.itemDelegate()
        configItemDelegate.finalizeEditor(editor)

        super(ConfigTreeView, self).closeEditor(editor, hint)


    def expandBranch(self, index=None, expanded=None):
        """ Expands or collapses the node at the index and all it's descendants.
            If expanded is True the nodes will be expanded, if False they will be collapsed, and if
            expanded is None the expanded attribute of each item is used.
            If parentIndex is None, the invisible root will be used (i.e. the complete forest will
            be expanded).
        """
        configModel = self.model()
        if index is None:
            #index = configTreeModel.createIndex()
            index = QtCore.QModelIndex()

        if index.isValid():
            if expanded is None:
                item = configModel.getItem(index)
                self.setExpanded(index, item.expanded)
            else:
                self.setExpanded(index, expanded)

        for rowNr in range(configModel.rowCount(index)):
            childIndex = configModel.index(rowNr, configModel.COL_NODE_NAME, parentIndex=index)
            self.expandBranch(index=childIndex, expanded=expanded)


    @property
    def autoReset(self):
        """ Indicates that the model will be (oartially) reset when the RTI or combo change
        """
        return self._configTreeModel.autoReset


    @autoReset.setter
    def autoReset(self, value):
        """ Indicates that the model will be (oartially) reset when the RTI or combo change
        """
        self._configTreeModel.autoReset = value


    @property
    def resetMode(self):
        """ Determines what is reset if autoReset is True (either axes or all settings)
        """
        return self._configTreeModel.resetMode


    @resetMode.setter
    def resetMode(self, value):
        """ Determines what is reset if autoReset is True (either axes or all settings)
        """
        self._configTreeModel.resetMode = value


    def resetAllSettings(self):
        """ Resets all settings
        """
        logger.debug("Resetting all settings")
        self._configTreeModel.resetAllSettings()


    def resetAllRanges(self):
        """ Resets all (axis/color/etc) range settings.
        """
        logger.debug("Resetting all range settings")
        self._configTreeModel.resetAllRanges()



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Abstract base classes for modeling data tree items for use in the ConfigTreeModel
"""
import enum
import logging, os

from argos.info import DEBUGGING, icons_directory
from argos.qt import Qt, QtCore, QtGui, QtWidgets, QtSlot
from argos.qt.treeitems import BaseTreeItem


logger = logging.getLogger(__name__)


###########
# Classes #
###########


class ResetMode(enum.Enum):
    All = "all"
    Ranges = "ranges"


DEFAULT_RESET_MODE = ResetMode.Ranges


class AbstractCti(BaseTreeItem):
    """ TreeItem for use in a ConfigTreeModel. (CTI = Config Tree Item)

        Abstract class. You must implement the _enforceDataType method that ensures the the data
        is stored internally in the correct type. Also, implement the createEditor method if you
        want you CTI to be editable (which is usually the case). If your CTI is not editable, make
        sure that valueColumnItemFlags does not return Qt.ItemIsEditable to prevent createEditor
        from being called by the delegate.

        Just like the BaseTreeItem every node has a name and a full path that is a slash-separated
        string of the full path leading from the root node to the current node. For instance, the
        path may be '/scale/x', the nodeName in that case is 'x'.

        CTIs are used to store a certain configuration value. It can be queried by the configValue
        property. The type of this value differs between descendants of AbstractCti, but a sub class
        should always return the same type. For example, the ColorCti.value always returns a
        QColor object. The displayValue returns the string representation for use in the tables;
        by default this returns: str(self.value)

        The underlying data is usually stored in that type as well but this is not necessarily so.
        A ChoiceCti, which represents a combo box, stores a list of choices and an index that is
        the actual choice made by the user. ChoiceCti.data contains the index while
        ChoiceCti.configValue returns: choices[index]. Note that the constructor expects the data
        as input parameter. The constructor calls the _enforceDataType method to convert the data
        to the correct type.

        The purpose of the defaultData is to reset a config data to its initial value when the
        user clicks on the reset button during editing.
    """
    def __init__(self, nodeName, defaultData, enabled=True, expanded=True):
        """ Constructor
            :param nodeName: name of this node (used to construct the node path).
            :param defaultData: default data to which the data can be reset by the reset button.
            :param enabled: True if the item is enabled
            :param expanded: True if the item is expanded
        """
        super(AbstractCti, self).__init__(nodeName=nodeName)

        self._defaultData = self._enforceDataType(defaultData)
        self._data = self.defaultData
        self._enabled = enabled
        self._expanded = expanded
        self._blockRefresh = False


    def finalize(self):
        """ Can be used to cleanup resources. Should be called explicitly.
            Recursively calls the _closeRecources method on all children and then on itself.
            Descendants should override _closeRecources to do the actual closing of resources.
        """
        for child in self.childItems:
            child.finalize()

        # In contrast to the BaseRti there is no close() method, _closeResources is directly called
        # because it only is needed when the config is deleted and catching exceptions makes no
        # sense.
        self._closeResources()


    def _closeResources(self):
        """ Can be overridden to close the underlying resources or disconnect signals.
            The default implementation does nothing.
            Is called by self.finalize when the cti is deleted. There is no corresponding
            _openResources; all resources are claimed in the constructor.
        """
        pass


    @property
    def debugInfo(self):
        """ Returns the string with debugging information
        """
        return ""


    def _dataToString(self, data):
        """ Conversion function used to convert the (default)data to the display value.
        """
        return str(data)

    @property
    def displayValue(self):
        """ Returns the string representation of data for use in the tree view.
            If a descendant overrides this, it should probably also override displayDefaultValue.
        """
        return self._dataToString(self.data)

    @property
    def displayDefaultValue(self):
        """ Returns the string representation of default data for use in the tree view.
        """
        return self._dataToString(self.defaultData)

    @property
    def configValue(self):
        """ Returns the configuration value of this item.
            By default this is the same as the underlying data but it can be overridden,
        """
        return self.data

    @property
    def data(self):
        """ Returns the data of this item.
        """
        return self._data

    @data.setter
    def data(self, data):
        """ Sets the data of this item.
            Does type conversion to ensure data is always of the correct type.
        """
        # Descendants should convert the data to the desired type here
        self._data = self._enforceDataType(data)

    @property
    def defaultData(self):
        """ Returns the default data of this item.
        """
        return self._defaultData

    @defaultData.setter
    def defaultData(self, defaultData):
        """ Sets the data of this item.
            Does type conversion to ensure default data is always of the correct type.
        """
        # Descendants should convert the data to the desired type here
        self._defaultData = self._enforceDataType(defaultData)

    def _enforceDataType(self, value):
        """ Converts data to the type of this CTI.
            Used by the setter to ensure that the data and defaultData have the correct type
        """
        raise NotImplementedError()

    @property
    def checkState(self):
        """ Returns how the checkbox for this cti should look like. Returns None for no checkbox.
            :rtype: Qt.CheckState or None
        """
        return None

    @checkState.setter
    def checkState(self, checkState):
        """ Allows the data to be set given a Qt.CheckState.
            Abstract method; you must override it if valueColumnItemFlags returns
            Qt.ItemIsUserCheckable
        """
        raise NotImplementedError()

    @property
    def enabled(self):
        """ Returns the enabled flag which determines if the user can interact with this item.
        """
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        """ Sets the enabled flag which determines if the user can interact with this item.
        """
        #logger.debug("Setting enabled = {:5} for {}".format(enabled, self))
        self._enabled = enabled


    @property
    def expanded(self):
        """ Returns the expanded flag which keeps
        """
        return self._expanded

    @expanded.setter
    def expanded(self, expanded):
        """ Keeps track if the config item is expanded.
            Needed to keep state between changing inspectors.
            Note, this can only be used as an initial expanded value but does not update the tree
            when it's constructed. Use ConfigTree.expand for that.
        """
        #logger.debug("Setting expanded = {!r:5} for {}".format(expanded, self))
        self._expanded = expanded


    def enableBranch(self, enabled):
        """ Sets the enabled member to True or False for a node and all it's children
        """
        self.enabled = enabled
        for child in self.childItems:
            child.enableBranch(enabled)


    def resetToDefault(self, resetChildren=True):
        """ Resets the data to the default data. By default the children will be reset as well
        """
        self.data = self.defaultData
        if resetChildren:
            for child in self.childItems:
                child.resetToDefault(resetChildren=True)


    def getRefreshBlocked(self):
        """ If True the calls to _refreshNodeFromTarget are blocked.
            This calls self.model.getRefreshBlocked(), so this value is always equal for all
            config tree items in the model
        """
        return self.model.getRefreshBlocked()


    def refreshFromTarget(self, level=0):
        """ Refreshes the configuration tree from the target it monitors (if present).
            Recursively call _refreshNodeFromTarget for itself and all children. Subclasses should
            typically override _refreshNodeFromTarget instead of this function.
            During updateTarget's execution refreshFromTarget is blocked to avoid loops.
        """
        if self.getRefreshBlocked():
            logger.debug("_refreshNodeFromTarget blocked")
            return

        if False and level == 0:
            logger.debug("refreshFromTarget: {}".format(self.nodePath))

        self._refreshNodeFromTarget()
        for child in self.childItems:
            child.refreshFromTarget(level=level + 1)


    def _refreshNodeFromTarget(self):
        """ Refreshes the configuration from the target it monitors (if present).
            The default implementation does nothing; subclasses can override it.
            During updateTarget's execution refreshFromTarget is blocked to avoid loops.
            Typically called from inspector.drawTarget
        """
        pass


    def updateTarget(self, level=0):
        """ Applies the configuration to the target it monitors (if present).
            Recursively call _updateTargetFromNode for itself and all children. Subclasses should
            typically override _updateTargetFromNode instead of this function.

            :param level: the level of recursion.
        """
        #if level == 0:
        #    logger.debug("updateTarget: {}".format(self.nodePath))

        self._updateTargetFromNode()
        for child in self.childItems:
            child.updateTarget(level = level + 1)


    def _updateTargetFromNode(self):
        """ Applies the configuration to the target it monitors (if present).
            The default implementation does nothing; subclasses can override it.
            During updateTarget's execution refreshFromTarget is blocked to avoid loops.
        """
        pass


    def logBranch(self, indent=0, level=logging.DEBUG):
        """ Logs the item and all descendants, one line per child
        """
        line = "{} ({})".format(self, self.configValue)
        if 0:
            print(indent * "    " + line)
        else:
            logger.log(level, indent * "    " + line)
        for childItems in self.childItems:
            childItems.logBranch(indent + 1, level=level)


    #################
    # serialization #
    #################


    def _nodeMarshall(self):
        """ Returns the non-recursive marshalled value of this CTI. Is called by marshall()
        """
        return self.data


    def marshall(self):
        """ Recursively retrieves values as a dictionary to be used for persistence.

            Typically descendants should override _nodeMarshall instead of this function.
        """
        nodeMar = self._nodeMarshall()
        assert not isinstance(nodeMar, dict), "Node marshall returns dict: {}".format(self)

        childDct = {}
        for childCti in self.childItems:
            childDct[childCti.nodeName] = childCti.marshall()

        if nodeMar is not None and childDct:
            res = {'_data': nodeMar, '_sub': childDct}
            return res
        elif childDct:
            return childDct
        else:
            return nodeMar


    def _nodeUnmarshall(self, data):
        """ Initializes itself non-recursively from data. Is called by unmarshall()
        """
        self.data = data


    def unmarshall(self, cfg):
        """ Initializes itself recursively from a config dict form the persistent settings.

            Typically descendants should override _nodeMarshall instead of this function.
        """
        if isinstance(cfg, dict):
            if '_data' in cfg and '_sub' in cfg:
                self._nodeUnmarshall(cfg['_data'])
                self.unmarshall(cfg['_sub'])
            else:
                for childName, childCfg in cfg.items():
                    try:
                        childCti = self.childByNodeName(childName)
                    except IndexError as _ex:
                        logger.warning("Unable to set values for: {}".format(childName))
                    else:
                        childCti.unmarshall(childCfg)
        else:
            self._nodeUnmarshall(cfg)



    # def marshall(self):
    #     """ Recursively retrieves values as a dictionary to be used for persistence.
    #
    #         Typically descendants should override _nodeMarshall instead of this function.
    #     """
    #
    #     if self.childItems:
    #         dct = {'_data': self._nodeMarshall()}
    #         dct['_sub'] = childDct = {}
    #         for childCti in self.childItems:
    #             childRes = childCti.marshall()
    #             childDct[childCti.nodeName] = childRes
    #     else:
    #         return self._nodeMarshall()
    #
    #     return dct

    ########################
    # Editor look and feel #
    ########################

    @property
    def valueColumnItemFlags(self):
        """ Returns Qt.ItemFlag enum that will be used for the value column in the config tree.
            These flags determine how the user can interact with the value column (e.g. can edit).

            Note that the ConfigTreeModel may override them: it will add the Qt.ItemIsEnabled and
            Qt.ItemIsSelectable to the flags.

            The base implementation of valueColumnItemFlags returns Qt.ItemIsEditable. Make sure to
            implement the createEditor abstract method if Qt.ItemIsEditable is included in the
            result.
        """
        return Qt.ItemIsEditable


    def createEditor(self, delegate, parent, option):
        """ Creates an editor (QWidget) for editing.
            It's parent will be set by the ConfigItemDelegate class that calls this method.

            :param delegate: the delegate that called this function
            :type  delegate: ConfigItemDelegate
            :param parent: The parent widget for the editor
            :type  parent: QWidget
            :param option: describes the parameters used to draw an item in a view widget.
            :type  option: QStyleOptionViewItem
        """
        raise NotImplementedError("createEditor not implemented while Qt.ItemIsEditable is set")



class AbstractCtiEditor(QtWidgets.QWidget):
    """ An editor for use in the ConfigTreeView (CTI).

        It consists of a horizontal collection of child widgets, the last of which is a reset
        button which will reset the config data to its default when clicked.

        You must implemented setData and getData, which pass the data from the
        QConfigItemDelegate to the editor and back.
    """
    def __init__(self, cti, delegate, subEditors=None, parent=None):
        """ Wraps the child widgets in a horizontal layout and appends a reset button.

            Maintains a reference to the ConfigTreeItem (cti) and to delegate, this last reference
            is so that we can command the delegate to commit and close the editor.

            The subEditors must be a list of QWidgets. Note that the sub editors do not yet have
            to be initialized with editor data since setData will be called by the delegate
            after construction. There it can be taken care of.
        """
        super(AbstractCtiEditor, self).__init__(parent=parent)

        # Prevent underlying table cell from being visible if the editor doesn't fill the cell
        self.setAutoFillBackground(True)

        self._subEditors = []
        self.delegate = delegate
        self.cti = cti

        # From the QAbstractItemDelegate.createEditor docs: The returned editor widget should have
        # Qt.StrongFocus; otherwise, QMouseEvents received by the widget will propagate to the view.
        # The view's background will shine through unless the editor paints its own background
        # (e.g., with setAutoFillBackground()).
        self.setFocusPolicy(Qt.StrongFocus)

        self.hBoxLayout = QtWidgets.QHBoxLayout()
        self.hBoxLayout.setContentsMargins(0, 0, 0, 0)
        self.hBoxLayout.setSpacing(0)
        self.setLayout(self.hBoxLayout)

        self.resetButton = QtWidgets.QToolButton()
        self.resetButton.setText("Reset")
        self.resetButton.setToolTip("Reset to default value.")
        self.resetButton.setIcon(QtGui.QIcon(os.path.join(icons_directory(), 'reset-l.svg')))
        self.resetButton.setFocusPolicy(Qt.NoFocus)
        self.resetButton.clicked.connect(self.resetEditorValue)
        self.hBoxLayout.addWidget(self.resetButton, alignment=Qt.AlignRight)

        self.cti.model.sigItemChanged.connect(self.modelItemChanged)

        for subEditor in (subEditors if subEditors is not None else []):
            self.addSubEditor(subEditor)


    def finalize(self):
        """ Called at clean up, when the editor is closed. Can be used to disconnect signals.
            This is often called after the client (e.g. the inspector) is updated. If you want to
            take action before the update, override prepareCommit instead.
            Be sure to call the finalize of the super class if you override this function.
        """
        for subEditor in self._subEditors:
            self.removeSubEditor(subEditor)

        self.cti.model.sigItemChanged.disconnect(self.modelItemChanged)
        self.resetButton.clicked.disconnect(self.resetEditorValue)
        self.cti = None # just to make sure it's not used again.
        self.delegate = None


    @QtSlot()
    def prepareCommit(self):
        """ Called just before the data is committed.
            Can be used to take action before the client (e.g. the inspector) is updated.
        """
        logger.debug("Committing data from: {}".format(self.cti.nodePath))


    def addSubEditor(self, subEditor, isFocusProxy=False):
        """ Adds a sub editor to the layout (at the right but before the reset button)
            Will add the necessary event filter to handle tabs and sets the strong focus so
            that events will not propagate to the tree view.

            If isFocusProxy is True the sub editor will be the focus proxy of the CTI.
        """
        self.hBoxLayout.insertWidget(len(self._subEditors), subEditor)
        self._subEditors.append(subEditor)

        subEditor.installEventFilter(self)
        subEditor.setFocusPolicy(Qt.StrongFocus)

        if isFocusProxy:
            self.setFocusProxy(subEditor)

        return subEditor


    def removeSubEditor(self, subEditor):
        """ Removes the subEditor from the layout and removes the event filter.
        """
        if subEditor is self.focusProxy():
            self.setFocusProxy(None)

        subEditor.removeEventFilter(self)
        self._subEditors.remove(subEditor)
        self.hBoxLayout.removeWidget(subEditor)


    def setData(self, value):
        """ Provides the editor widget with a data to manipulate.
            Value originates from the ConfigTreeModel.data(role=QEditRole).
        """
        raise NotImplementedError()


    def getData(self):
        """ Gets data from the editor widget.
            Should return a value that can be set into the ConfigTreeModel with the QEditRole.
        """
        raise NotImplementedError()


    def eventFilter(self, watchedObject, event):
        """ Calls commitAndClose when the tab and back-tab are pressed.
            This is necessary because, normally the event filter of QStyledItemDelegate does this
            for us. However, that event filter works on this object, not on the sub editor.
        """
        if event.type() == QtCore.QEvent.KeyPress:
            key = event.key()
            if key in (Qt.Key_Tab, Qt.Key_Backtab):
                self.commitAndClose()
                return True
            else:
                return False

        return super(AbstractCtiEditor, self).eventFilter(watchedObject, event)


    @QtSlot(BaseTreeItem)
    def modelItemChanged(self, cti):
        """ Called when the an Config Tree Item (CTI) in the model has changed.

            If the CTI is a different one than the CTI that belongs to this editor, the editor
            is closed. This can happen if the user has checked a checkbox. Qt does not close other
            editors in the view in that case, so this is why we do it here.

            If the cti parameter is the CTI belonging to this editor, nothing is done. We don't
            close the editor because the user may want to continue editing.
        """
        if cti is not self.cti:
            logger.debug("Another config tree item has changed: {}. Closing editor for {}"
                         .format(cti, self.cti))
            self.delegate.closeEditor.emit(self, QtWidgets.QAbstractItemDelegate.NoHint) # CLOSES SELF!
        else:
            logger.debug("Cti of this editor has changed: {}".format(cti))


    @QtSlot()
    def commitAndClose(self):
        """ Commits the data of the sub editor and instructs the delegate to close this ctiEditor.

            The delegate will emit the closeEditor signal which is connected to the closeEditor
            method of the ConfigTreeView class. This, in turn will, call the finalize method of
            this object so that signals can be disconnected and resources can be freed. This is
            complicated but I don't see a simpler solution.
        """
        if self.delegate:
            self.delegate.commitData.emit(self)
            self.delegate.closeEditor.emit(self, QtWidgets.QAbstractItemDelegate.NoHint) # CLOSES SELF!
        else:
            # QAbstractItemView.closeEditor is sometimes called directly, without the
            # QAbstractItemDelegate.closeEditor signal begin emitted, e.g when the currentItem
            # changes. Therefore the commitAndClose method can be called twice, if we call it
            # explicitly as well (e.g. in FontCtiEditor.execFontDialog(). We guard against this.
            logger.debug("AbstractCtiEditor.commitAndClose: editor already closed (ignored).")


    @QtSlot(bool)
    def resetEditorValue(self, checked=False):
        """ Resets the editor to the default value. Also resets the children.
        """
        # Block all signals to prevent duplicate inspector updates.
        # No need to restore, the editors will be deleted after the reset.
        for subEditor in self._subEditors:
            subEditor.blockSignals(True)

        self.cti.resetToDefault(resetChildren=True)
        # This will commit the children as well.
        self.setData(self.cti.defaultData)
        self.commitAndClose()


    def paintEvent(self, event):
        """ Reimplementation of paintEvent to allow for style sheets
            See: http://qt-project.org/wiki/How_to_Change_the_Background_Color_of_QWidget
        """
        opt = QtWidgets.QStyleOption()
        opt.initFrom(self)
        painter = QtGui.QPainter(self)
        self.style().drawPrimitive(QtWidgets.QStyle.PE_Widget, opt, painter, self)
        painter.end()




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Data repository functionality
"""
import logging

from argos.qt import Qt, QtCore, QtSlot
from argos.qt.treemodels import BaseTreeModel
from argos.config.abstractcti import ResetMode, DEFAULT_RESET_MODE
from argos.config.groupcti import MainGroupCti
from argos.utils.cls import type_name, check_class


logger = logging.getLogger(__name__)

class ConfigTreeModel(BaseTreeModel):
    """ An implementation QAbstractItemModel that offers access to configuration data for QTreeViews.
        The underlying data is stored as config tree items (AbstractCti descendants)
    """
    HEADERS = ["Name", "Path", "Value", "Default value", "Item class", "Debug info"]
    (COL_NODE_NAME, COL_NODE_PATH,
     COL_VALUE, COL_DEF_VALUE,
     COL_CTI_TYPE, COL_DEBUG) = range(len(HEADERS))

    COL_DECORATION = COL_VALUE   # Column number that contains the decoration.

    INVISIBLE_ROOT_NAME = '<invisible-root>'

    def __init__(self, parent=None):
        """ Constructor
        """
        super(ConfigTreeModel, self).__init__(parent=parent)

        self._autoReset = True
        self._resetMode = DEFAULT_RESET_MODE

        self._invisibleRootTreeItem = MainGroupCti(self.INVISIBLE_ROOT_NAME)
        self._invisibleRootTreeItem.model = self
        #self.dataChanged.connect(self.debug)
        self._refreshBlocked = False


    def flags(self, index):
        """ Returns the item flags for the given index.
        """
        if not index.isValid():
            return 0

        cti = self.getItem(index)

        result = Qt.ItemIsSelectable

        if cti.enabled:
            result |= Qt.ItemIsEnabled

        if index.column() == self.COL_VALUE:
            result |= cti.valueColumnItemFlags

        return result


    def itemData(self, treeItem, column, role=Qt.DisplayRole):
        """ Returns the data stored under the given role for the item.
        """
        if role == Qt.DisplayRole:
            if column == self.COL_NODE_NAME:
                return treeItem.nodeName
            elif column == self.COL_NODE_PATH:
                return treeItem.nodePath
            elif column == self.COL_VALUE:
                return treeItem.displayValue
            elif column == self.COL_DEF_VALUE:
                return treeItem.displayDefaultValue
            elif column == self.COL_CTI_TYPE:
                return type_name(treeItem)
            elif column == self.COL_DEBUG:
                return treeItem.debugInfo
            else:
                raise ValueError("Invalid column: {}".format(column))

        elif role == Qt.EditRole:
            if column == self.COL_VALUE:
                return treeItem.data
            else:
                raise ValueError("Invalid column: {}".format(column))

        elif role == Qt.ToolTipRole:
            if column == self.COL_NODE_NAME or column == self.COL_NODE_PATH:
                return treeItem.nodePath
            elif column == self.COL_VALUE:
                # Give Access to exact values. In particular in scientific-notation spin boxes
                return repr(treeItem.configValue)
            elif column == self.COL_DEF_VALUE:
                return treeItem.displayDefaultValue
            elif column == self.COL_CTI_TYPE:
                return type_name(treeItem)
            elif column == self.COL_DEBUG:
                return treeItem.debugInfo
            else:
                return None

        elif role == Qt.CheckStateRole:
            if column != self.COL_VALUE:
                # The CheckStateRole is called for each cell so return None here.
                return None
            else:
                return treeItem.checkState
        else:
            return super(ConfigTreeModel, self).itemData(treeItem, column, role=role)


    def setItemData(self, treeItem, column, value, role=Qt.EditRole):
        """ Sets the role data for the item at index to value.
        """
        if role == Qt.CheckStateRole:
            if column != self.COL_VALUE:
                return False
            else:
                logger.debug("Setting check state (col={}): {!r}".format(column, value))
                treeItem.checkState = value
                return True

        elif role == Qt.EditRole:
            if column != self.COL_VALUE:
                return False
            else:
                logger.debug("Set Edit value (col={}): {!r}".format(column, value))
                treeItem.data = value
                return True
        else:
            raise ValueError("Unexpected edit role: {}".format(role))


    def setExpanded(self, index, expanded):
        """ Expands the model item specified by the index.
            Overridden from QTreeView to make it persistent (between inspector changes).
        """
        if index.isValid():
            item = self.getItem(index)
            item.expanded = expanded
            #logger.debug("Setting expanded = {} for {}".format(expanded, item))


    def expand(self, index):
        """ Expands the model item specified by the index.
            Overridden from QTreeView to make it persistent (between inspector changes).
        """
        self.setExpanded(index, True)


    def collapse(self, index):
        """ Expands the model item specified by the index.
            Overridden from QTreeView to make it persistent (between inspector changes).
        """
        self.setExpanded(index, False)


    def indexTupleFromItem(self, treeItem): # TODO: move to BaseTreeItem?
        """ Return (first column model index, last column model index) tuple for a configTreeItem
        """
        if not treeItem:
            return (QtCore.QModelIndex(), QtCore.QModelIndex())

        if not treeItem.parentItem: # TODO: only necessary because of childNumber?
            return (QtCore.QModelIndex(), QtCore.QModelIndex())

        # Is there a bug in Qt in QStandardItemModel::indexFromItem?
        # It passes the parent in createIndex. TODO: investigate

        row =  treeItem.childNumber()
        return (self.createIndex(row, 0, treeItem),
                self.createIndex(row, self.columnCount() - 1, treeItem))


    def emitDataChanged(self, treeItem):  # TODO: move to BaseTreeItem?
        """ Emits the data changed for the model indices (all columns) for this treeItem
        """
        indexLeft, indexRight = self.indexTupleFromItem(treeItem)
        self.dataChanged.emit(indexLeft, indexRight)


    def getRefreshBlocked(self):
        """ If set the configuration should not be updated.
            This setting is part of the model so that is shared by all CTIs.
        """
        return self._refreshBlocked


    def setRefreshBlocked(self, blocked):
        """ Set to True to indicate that set the configuration should not be updated.
            This setting is part of the model so that is shared by all CTIs.
            Returns the old value.
        """
        wasBlocked = self._refreshBlocked
        logger.debug("Setting refreshBlocked from {} to {}".format(wasBlocked, blocked))
        self._refreshBlocked = blocked
        return wasBlocked


    @property
    def autoReset(self):
        """ Indicates that the model will be (oartially) reset when the RTI or combo change
        """
        return self._autoReset


    @autoReset.setter
    def autoReset(self, value):
        """ Indicates that the model will be (oartially) reset when the RTI or combo change
        """
        self._autoReset = value


    @property
    def resetMode(self):
        """ Determines what is reset if autoReset is True (either axes or all settings)
        """
        return self._resetMode


    @resetMode.setter
    def resetMode(self, value):
        """ Determines what is reset if autoReset is True (either axes or all settings)
        """
        check_class(value, ResetMode)
        self._resetMode = value


    def resetAllSettings(self):
        """ Resets all items in the tree to their default
        """
        logger.debug("Resetting all settings")
        self.invisibleRootTreeItem.resetToDefault(resetChildren=True)
        self.emitDataChanged(self.invisibleRootTreeItem)     # Updates the tree
        self.sigItemChanged.emit(self.invisibleRootTreeItem) # Updates the inspector


    def resetAllRanges(self):
        """ Resets all (axit/color/ect) range items in the tree to their default
        """
        logger.debug("Resetting all settings")
        self.invisibleRootTreeItem.resetRangesToDefault()
        self.emitDataChanged(self.invisibleRootTreeItem)     # Updates the tree
        self.sigItemChanged.emit(self.invisibleRootTreeItem) # Updates the inspector




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos.  If not, see <http://www.gnu.org/licenses/>.

""" Create batch file wrappers so that Argos can be called from outside a (conda) environment.

    An alternative may be to use exec-wrappers.
        https://pypi.org/project/exec-wrappers/
        https://stackoverflow.com/questions/42025173/best-way-to-execute-a-python-script-in-a-given-conda-environment
"""
from __future__ import print_function

import argparse
import logging
import os.path
import sys

from argos.info import DEBUGGING, VERSION
from argos.utils.logs import make_log_format, initLogging
from argos.utils.misc import remove_process_serial_number

APP_NAME = 'argos_make_wrappers'

logger = logging.getLogger("argos." + APP_NAME)

logging.basicConfig(level='INFO', stream=sys.stderr, format=make_log_format())


def eprint(*args, **kwargs):
    """ Print to stderr
    """
    print(*args, file=sys.stderr, **kwargs)



def exitIfNotExists(fileName, prefix="File", postFix=""):
    """ Exits the program if the fileName does not exist.
    """
    if not os.path.exists(fileName):
        eprint("{} doesn't exist: '{}'. {}".format(prefix, fileName, postFix))
        sys.exit(1)


def exitIfNotAFile(fileName, prefix="File", postFix=""):
    """ Exits the program if the fileName does not exist or is a directory
    """
    exitIfNotExists(fileName, prefix, postFix)
    if os.path.isdir(fileName):
        eprint("{} is a directory: '{}. {}'".format(prefix, fileName, postFix))
        sys.exit(1)


def exitIfNotADir(fileName, prefix="File", postFix=""):
    """ Exits the program if the fileName does not exist or is NOT a directory
    """
    exitIfNotExists(fileName, prefix, postFix)
    if not os.path.isdir(fileName):
        eprint("{} is not a directory: '{}. {}'".format(prefix, fileName, postFix))
        sys.exit(1)


# C:\Users\kenter\Miniconda3\Scripts\conda.exe run -n argosdev C:\Users\kenter\Miniconda3\envs\argosdev\Scripts\argos.exe %*
TEMPLATE = r"""@rem Batch file for starting Argos from outside this Anaconda environment.
@rem Created with {appName}

@{condaExecutable} run -n {condaEnvironment} {argosExecutable} %*
"""

def writeScript(fileName, text, dryRun=False):
    """ Creates and writes the text to a wrapper script named fileName.

        If dryRun is True, no files are created but the scripts contents are printed on stdout.
    """
    if dryRun:
        logger.info("Simulate creating argos wrapper: {}".format(fileName))
        print("==== {} ====".format(fileName))
        print(text)
        print()
    else:
        logger.info("Creating argos wrapper: {}".format(fileName))
        with open(fileName, 'w') as fileOut:
            print(text, file=fileOut)


def makeCondaWrappers(dryRun=False, condaEnvironment=None, condaPrefix=None, condaExecutable=None):
    """ Creates wrapper batch scripts for Anaconda.
    """
    if dryRun:
        logger.warning("Not creating wrappers but printing to stdout instead.")

    logger.debug("Anaconda environment variables found:")
    for key, value in os.environ.items():
        if key.startswith('CONDA'):
            logger.debug("    {:25s} = {}".format(key, value))

    if condaEnvironment:
        logger.debug("Using conda environment name specified with the command line argument.")
    else:
        logger.debug("Using conda environment name from the CONDA_DEFAULT_ENV environment variable.")
        try:
            condaEnvironment = os.environ['CONDA_DEFAULT_ENV']
        except KeyError:
            eprint("CONDA_DEFAULT_ENV environment variable not found.")
            sys.exit(1)
    logger.info("Conda environment: '{}'".format(condaEnvironment))

    if condaPrefix:
        logger.debug("Using conda prefix specified with the command line argument.")
    else:
        logger.debug("Using conda prefix from the CONDA_PREFIX environment variable.")
        try:
            condaPrefix = os.environ['CONDA_PREFIX']
        except KeyError:
            eprint("CONDA_PREFIX environment variable not found.")
            sys.exit(1)
    logger.info("Conda prefix: '{}'".format(condaPrefix))
    exitIfNotADir(condaPrefix, "\nThe conda prefix",
                  "\n\nYou can use the --conda-prefix command line argument to specify it.")

    if condaEnvironment != "base" and not condaPrefix.endswith(condaEnvironment):
        logger.warning("Conda prefix ({}) does not end with conda environment ({})."
                       .format(condaPrefix, condaEnvironment))

    if condaExecutable:
        logger.debug("Using conda executable specified with the command line argument.")
    else:
        logger.debug("Using conda executable from the CONDA_EXE environment variable.")
        try:
            condaExecutable = os.environ['CONDA_EXE']
        except KeyError:
            eprint("CONDA_EXE environment variable not found.")
            sys.exit(1)

    condaExecutable = os.path.abspath(condaExecutable)
    logger.info("Conda executable: '{}'".format(condaExecutable))
    exitIfNotAFile(condaExecutable, "\nThe conda executable",
                   "\n\nYou can use the --conda-exe command line argument to specify it.")

    condaScriptsDir = os.path.join(condaPrefix, "Scripts")
    exitIfNotADir(condaScriptsDir, "\nThe conda Scripts dir")

    for argosBaseName in ['argos', 'argosw']:

        argosBasePath = os.path.normpath(os.path.join(condaScriptsDir, argosBaseName))
        argosExecutable = "{}.exe".format(argosBasePath)
        logger.debug("Argos executable: {}".format(argosExecutable))
        if not os.path.exists(argosExecutable):
            logger.warning("Argos executable does not exist: '{}'".format(argosExecutable))

        argosWrapper = "{}.bat".format(argosBasePath)

        text = TEMPLATE.format(
            appName=APP_NAME,
            condaExecutable=condaExecutable,
            argosExecutable=argosExecutable,
            condaEnvironment=condaEnvironment)

        writeScript(argosWrapper, text, dryRun=dryRun)



def main():
    """ Create batch file wrappers so that Argos can be called from outside a (conda) environment.
    """
    aboutStr = "{} version: {}".format("argos_make_wrappers", VERSION)
    parser = argparse.ArgumentParser(description = __doc__)

    parser.add_argument('-d', '--dry-run', dest='dryRun', action = 'store_true',
        help="""If set, no wrapper files are created but their contents is printed on stdout.""")

    parser.add_argument('--env', '--conda-env', dest='condaEnvironment',
        help="""Anaconda environment name, e.g. "base". If not given, the script tries to find it
                with the use of the environment variables in the current Anaconda environment.""")

    parser.add_argument('--prefix', '--conda-prefix', dest='condaPrefix',
        help=r"""Path to the Anaconda environment, e.g. "C:\Users\my_name\Miniconda3".
                If not given, the script tries to find it with the use of the environment variables
                in the current Anaconda environment.""")

    parser.add_argument('--exe', '--conda-exe', dest='condaExecutable',
        help="""Full path to the conda.exe executable. If not given, the script tries to find it
                with the use of the environment variables in the current Anaconda environment.""")

    parser.add_argument('-l', '--log-level', dest='log_level', default='info',
        help="Log level. If set, only log messages with a level higher or equal than this will be "
             "printed to screen (stderr). Overrides the log level of the StreamHandlers in the "
             "--log-config file. Does not alter the log level of log handlers that write to a "
             "file.",
        choices=('debug', 'info', 'warning', 'error', 'critical'))

    parser.add_argument('--version', action = 'store_true',
        help="Prints the program version and exits")

    args = parser.parse_args(remove_process_serial_number(sys.argv[1:]))

    initLogging(None, args.log_level)

    if args.version:
        print(aboutStr)
        sys.exit(0)

    logger.info(aboutStr)
    logger.debug("argv: {}".format(sys.argv))
    logger.debug("PID: {}".format(os.getpid()))

    makeCondaWrappers(
        dryRun=args.dryRun,
        condaExecutable=args.condaExecutable,
        condaPrefix=args.condaPrefix,
        condaEnvironment=args.condaEnvironment
    )

    logger.info('Done {}'.format(APP_NAME))

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" PyQtGraph 2D image plot
"""
from __future__ import division, print_function

import os
import logging, math
import numpy as np
import pyqtgraph as pg

from functools import partial
from collections import OrderedDict

from argos.info import DEBUGGING
from argos.config.boolcti import BoolCti, BoolGroupCti
from argos.config.choicecti import ChoiceCti
from argos.config.groupcti import MainGroupCti
from argos.inspector.abstract import AbstractInspector, InvalidDataError, UpdateReason
from argos.inspector.pgplugins.colorbar import ArgosColorLegendItem
from argos.inspector.pgplugins.pgctis import (
    X_AXIS, Y_AXIS, BOTH_AXES, NO_LABEL_STR, defaultAutoRangeMethods, PgAxisLabelCti,
    PgAxisCti, PgAxisFlipCti, PgAspectRatioCti, PgAxisRangeCti, PgGridCti,
    PgColorMapCti, PgColorLegendCti, PgColorLegendLabelCti, PgShowHistCti,
    PgShowDragLinesCti, setXYAxesAutoRangeOn, PgPlotDataItemCti)
from argos.inspector.pgplugins.pgplotitem import ArgosPgPlotItem
from argos.qt import Qt, QtCore, QtGui, QtSlot

from argos.utils.cls import array_has_real_numbers, check_class, is_an_array, to_string
from argos.utils.cls import array_kind_label
from argos.utils.defs import RIGHT_ARROW
from argos.utils.masks import (ArrayWithMask, replaceMaskedValueWithFloat,
                               nanPercentileOfSubsampledArrayWithMask)

logger = logging.getLogger(__name__)

ROW_TITLE,    COL_TITLE    = 0, 0  # colspan = 3
ROW_COLOR,    COL_COLOR    = 1, 2  # rowspan = 2
ROW_HOR_LINE, COL_HOR_LINE = 1, 0
ROW_IMAGE,    COL_IMAGE    = 2, 0
ROW_VER_LINE, COL_VER_LINE = 2, 1
ROW_PROBE,    COL_PROBE    = 3, 0  # colspan = 2


def calcPgImagePlot2dDataRange(pgImagePlot2d, percentage, crossPlot, subsample):
    """ Calculates the range from the inspectors' sliced array. Discards percentage of the minimum
        and percentage of the maximum values of the inspector.slicedArray

        :param pgImagePlot2d: the range methods will work on (the sliced array) of this inspector.
        :param float percentage: percentage that will be discarded.
        :param bool crossPlot: if None, the range will be calculated from the entire sliced array,
            if "horizontal" or "vertical" the range will be calculated from the data under the
            horizontal or vertical cross hairs.
            If the cursor is outside the image, there is no valid data under the cross-hair and
            the range will be determined from the sliced array as a fall back.
        :param bool subsample: if True, the image will be subsampled (to 200 by 200) before
            calculating the range. This to improve performance by large images.
    """
    check_class(pgImagePlot2d.slicedArray, ArrayWithMask) # sanity check

    if crossPlot is None:
        array = pgImagePlot2d.slicedArray  # the whole image

    elif crossPlot == 'horizontal':
        if pgImagePlot2d.crossPlotRow is not None:
            array = pgImagePlot2d.slicedArray[pgImagePlot2d.crossPlotRow, :]
        else:
            array = pgImagePlot2d.slicedArray # fall back on complete sliced array

    elif crossPlot == 'vertical':
        if pgImagePlot2d.crossPlotCol is not None:
            array = pgImagePlot2d.slicedArray[:, pgImagePlot2d.crossPlotCol]
        else:
            array = pgImagePlot2d.slicedArray # fall back on complete sliced array
    else:
        raise ValueError("crossPlot must be: None, 'horizontal' or 'vertical', got: {}"
                         .format(crossPlot))

    return nanPercentileOfSubsampledArrayWithMask(array, (percentage, 100 - percentage), subsample)


def crossPlotAutoRangeMethods(pgImagePlot2d, crossPlot, intialItems=None):
    """ Creates an ordered dict with autorange methods for an PgImagePlot2d inspector.

        :param pgImagePlot2d: the range methods will work on (the sliced array) of this inspector.
        :param crossPlot: if None, the range will be calculated from the entire sliced array,
            if "horizontal" or "vertical" the range will be calculated from the data under the
            horizontal or vertical cross hairs
        :param intialItems: will be passed on to the  OrderedDict constructor.
    """
    rangeFunctions = OrderedDict({} if intialItems is None else intialItems)

    # If crossPlot is "horizontal" or "vertical" make functions that determine the range from the
    # data at the cross hair.
    if crossPlot:
        rangeFunctions['cross all data'] = partial(calcPgImagePlot2dDataRange, pgImagePlot2d,
                                                   0.0, crossPlot)
        for percentage in [0.1, 0.2, 0.5, 1, 2, 5, 10, 20]:
            label = "cross discard {}%".format(percentage)
            rangeFunctions[label] = partial(calcPgImagePlot2dDataRange, pgImagePlot2d,
                                            percentage, crossPlot)

    # Always add functions that determine the data from the complete sliced array.
    for percentage in [0.1, 0.2, 0.5, 1, 2, 5, 10, 20]:
        rangeFunctions['image all data'] = partial(calcPgImagePlot2dDataRange, pgImagePlot2d,
                                                   0.0, None)

        label = "image discard {}%".format(percentage)
        rangeFunctions[label] = partial(calcPgImagePlot2dDataRange, pgImagePlot2d,
                                        percentage, None)
    return rangeFunctions



class PgImagePlot2dCti(MainGroupCti):
    """ Configuration tree item for a PgImagePlot2dCti inspector
    """
    def __init__(self, pgImagePlot2d, nodeName):
        """ Constructor

            Maintains a link to the target pgImagePlot2d inspector, so that changes in the
            configuration can be applied to the target by simply calling the apply method.
            Vice versa, it can connect signals to the target.
        """
        super(PgImagePlot2dCti, self).__init__(nodeName)
        check_class(pgImagePlot2d, PgImagePlot2d)
        self.pgImagePlot2d = pgImagePlot2d
        imagePlotItem = self.pgImagePlot2d.imagePlotItem
        viewBox = imagePlotItem.getViewBox()

        self.insertChild(
            ChoiceCti('title', 0, editable=True,
                      configValues=["{base-name} -- {name} {slices}",
                                    "{name} {slices}", "{path} {slices}"]))

        #### Axes ####

        self.aspectLockedCti = self.insertChild(PgAspectRatioCti(viewBox))

        self.xAxisCti = self.insertChild(PgAxisCti('x-axis'))
        self.xAxisCti.insertChild(
            PgAxisLabelCti(imagePlotItem, 'bottom', self.pgImagePlot2d.collector,
                           defaultData=1,
                           configValues=[NO_LABEL_STR, "{x-dim} [index]"]))
        self.xFlippedCti = self.xAxisCti.insertChild(PgAxisFlipCti(viewBox, X_AXIS))
        self.xAxisRangeCti = self.xAxisCti.insertChild(PgAxisRangeCti(viewBox, X_AXIS))

        self.yAxisCti = self.insertChild(PgAxisCti('y-axis'))
        self.yAxisCti.insertChild(
            PgAxisLabelCti(imagePlotItem, 'left', self.pgImagePlot2d.collector,
                           defaultData=1,
                           configValues=[NO_LABEL_STR, "{y-dim} [index]"]))
        self.yFlippedCti = self.yAxisCti.insertChild(
            PgAxisFlipCti(viewBox, Y_AXIS, defaultData=True))
        self.yAxisRangeCti = self.yAxisCti.insertChild(PgAxisRangeCti(viewBox, Y_AXIS))

        #### Color scale ####

        self.colorCti = self.insertChild(PgAxisCti('color scale'))

        self.colorCti.insertChild(PgColorLegendLabelCti(
            pgImagePlot2d.colorLegendItem, self.pgImagePlot2d.collector, defaultData=1,
            configValues=[NO_LABEL_STR, "{name} {unit}", "{path} {unit}",
                          "{name}", "{path}", "{raw-unit}"]))

        self.colorCti.insertChild(PgColorMapCti(self.pgImagePlot2d.colorLegendItem))

        self.showHistCti = self.colorCti.insertChild(
            PgShowHistCti(pgImagePlot2d.colorLegendItem))
        self.showDragLinesCti = self.colorCti.insertChild(
            PgShowDragLinesCti(pgImagePlot2d.colorLegendItem))

        colorAutoRangeFunctions = defaultAutoRangeMethods(self.pgImagePlot2d)
        self.colorLegendCti = self.colorCti.insertChild(
            PgColorLegendCti(pgImagePlot2d.colorLegendItem, colorAutoRangeFunctions,
                             nodeName="range"))

        # If True, the image is automatically downsampled to match the screen resolution. This
        # improves performance for large images and reduces aliasing. If autoDownsample is not
        # specified, then ImageItem will choose whether to downsample the image based on its size.
        self.autoDownSampleCti = self.insertChild(BoolCti('auto down sample', True))
        self.zoomModeCti = self.insertChild(BoolCti('rectangle zoom mode', False))

        ### Probe and cross-hair plots ###

        self.probeCti = self.insertChild(BoolCti('show probe', True))
        self.crossPlotGroupCti = self.insertChild(BoolGroupCti('cross-hair', expanded=False))
        self.crossPenCti = self.crossPlotGroupCti.insertChild(PgPlotDataItemCti(expanded=False))

        self.horCrossPlotCti = self.crossPlotGroupCti.insertChild(
            BoolCti('horizontal', False, expanded=False))

        self.horCrossPlotCti.insertChild(PgGridCti(pgImagePlot2d.horCrossPlotItem))
        self.horCrossPlotRangeCti = self.horCrossPlotCti.insertChild(
            PgAxisRangeCti(
                self.pgImagePlot2d.horCrossPlotItem.getViewBox(), Y_AXIS, nodeName="data range",
                autoRangeFunctions=crossPlotAutoRangeMethods(self.pgImagePlot2d, "horizontal")))

        self.verCrossPlotCti = self.crossPlotGroupCti.insertChild(
            BoolCti('vertical', False, expanded=False))
        self.verCrossPlotCti.insertChild(PgGridCti(pgImagePlot2d.verCrossPlotItem))
        self.verCrossPlotRangeCti = self.verCrossPlotCti.insertChild(
            PgAxisRangeCti(
                self.pgImagePlot2d.verCrossPlotItem.getViewBox(), X_AXIS, nodeName="data range",
                autoRangeFunctions=crossPlotAutoRangeMethods(self.pgImagePlot2d, "vertical")))

        # Connect signals.

        # Use a queued connect to schedule the reset after current events have been processed.
        self.pgImagePlot2d.colorLegendItem.sigResetColorScale.connect(
            self.colorLegendCti.setAutoRangeOn, type=Qt.QueuedConnection)
        self.pgImagePlot2d.imagePlotItem.sigResetAxis.connect(
            self.setImagePlotAutoRangeOn, type=Qt.QueuedConnection)
        self.pgImagePlot2d.horCrossPlotItem.sigResetAxis.connect(
            self.setHorCrossPlotAutoRangeOn, type=Qt.QueuedConnection)
        self.pgImagePlot2d.verCrossPlotItem.sigResetAxis.connect(
            self.setVerCrossPlotAutoRangeOn, type=Qt.QueuedConnection)

        # Also update axis auto range tree items when linked axes are resized
        horCrossViewBox = self.pgImagePlot2d.horCrossPlotItem.getViewBox()
        horCrossViewBox.sigRangeChangedManually.connect(self.xAxisRangeCti.setAutoRangeOff)
        verCrossViewBox = self.pgImagePlot2d.verCrossPlotItem.getViewBox()
        verCrossViewBox.sigRangeChangedManually.connect(self.yAxisRangeCti.setAutoRangeOff)


    def _closeResources(self):
        """ Disconnects signals.
            Is called by self.finalize when the cti is deleted.
        """
        self.pgImagePlot2d.colorLegendItem.sigResetColorScale.disconnect(
            self.colorLegendCti.setAutoRangeOn)

        verCrossViewBox = self.pgImagePlot2d.verCrossPlotItem.getViewBox()
        verCrossViewBox.sigRangeChangedManually.disconnect(self.yAxisRangeCti.setAutoRangeOff)
        horCrossViewBox = self.pgImagePlot2d.horCrossPlotItem.getViewBox()
        horCrossViewBox.sigRangeChangedManually.disconnect(self.xAxisRangeCti.setAutoRangeOff)

        self.pgImagePlot2d.verCrossPlotItem.sigResetAxis.disconnect(self.setVerCrossPlotAutoRangeOn)
        self.pgImagePlot2d.horCrossPlotItem.sigResetAxis.disconnect(self.setHorCrossPlotAutoRangeOn)
        self.pgImagePlot2d.imagePlotItem.sigResetAxis.disconnect(self.setImagePlotAutoRangeOn)


    def setImagePlotAutoRangeOn(self, axisNumber):
        """ Sets the image plot's auto-range on for the axis with number axisNumber.

            :param axisNumber: 0 (X-axis), 1 (Y-axis), 2, (Both X and Y axes).
        """
        setXYAxesAutoRangeOn(self, self.xAxisRangeCti, self.yAxisRangeCti, axisNumber)


    def setHorCrossPlotAutoRangeOn(self, axisNumber):
        """ Sets the horizontal cross-hair plot's auto-range on for the axis with number axisNumber.

            :param axisNumber: 0 (X-axis), 1 (Y-axis), 2, (Both X and Y axes).
        """
        setXYAxesAutoRangeOn(self, self.xAxisRangeCti, self.horCrossPlotRangeCti, axisNumber)


    def setVerCrossPlotAutoRangeOn(self, axisNumber):
        """ Sets the vertical cross-hair plot's auto-range on for the axis with number axisNumber.

            :param axisNumber: 0 (X-axis), 1 (Y-axis), 2, (Both X and Y axes).
        """
        setXYAxesAutoRangeOn(self, self.verCrossPlotRangeCti, self.yAxisRangeCti, axisNumber)


    def resetRangesToDefault(self):
        """ Resets range settings to the default data.
        """
        self.xAxisRangeCti.autoRangeCti.data = True
        self.yAxisRangeCti.autoRangeCti.data = True
        self.colorLegendCti.autoRangeCti.data = True
        self.horCrossPlotRangeCti.autoRangeCti.data = True
        self.verCrossPlotRangeCti.autoRangeCti.data = True



class PgImagePlot2d(AbstractInspector):
    """ Draws an image plot of a two-dimensional array (slice).

        Plotting is done with the PyQtGraph package. See www.pyqtgraph.org.
    """

    def __init__(self, collector, parent=None):
        """ Constructor. See AbstractInspector constructor for parameters.
        """
        super(PgImagePlot2d, self).__init__(collector, parent=parent)

        # Ensure that only a white background is visible when self.graphicsLayoutWidget is hidden.
        self.contentsWidget.setStyleSheet('background: white;')

        # The sliced array is kept in memory. This may be different per inspector, e.g. 3D
        # inspectors may decide that this uses to much memory. The slice is therefor not stored
        # in the collector.
        self.slicedArray = None

        self.titleLabel = pg.LabelItem('title goes here...')

        # The image item
        self.imagePlotItem = ArgosPgPlotItem()
        self.viewBox = self.imagePlotItem.getViewBox()
        self.viewBox.disableAutoRange(BOTH_AXES)

        self.imageItem = pg.ImageItem()
        self.imageItem.setPos(-0.5, -0.5) # Center on pixels (see pg.ImageView.setImage source code)
        self.imagePlotItem.addItem(self.imageItem)

        self.colorLegendItem = ArgosColorLegendItem(self.imageItem)

        # Probe and cross hair plots
        self.crossPlotRow = None # the row coordinate of the cross hair. None if no cross hair.
        self.crossPlotCol = None # the col coordinate of the cross hair. None if no cross hair.
        self.horCrossPlotItem = ArgosPgPlotItem()
        self.verCrossPlotItem = ArgosPgPlotItem()
        self.horCrossPlotItem.setXLink(self.imagePlotItem)
        self.verCrossPlotItem.setYLink(self.imagePlotItem)
        self.horCrossPlotItem.setLabel('left', ' ')
        self.verCrossPlotItem.setLabel('bottom', ' ')
        self.horCrossPlotItem.showAxis('top', True)
        self.horCrossPlotItem.showAxis('bottom', False)
        self.verCrossPlotItem.showAxis('right', True)
        self.verCrossPlotItem.showAxis('left', False)

        self.crossPen = pg.mkPen("#BFBFBF")
        self.crossShadowPen = pg.mkPen([0, 0, 0, 100], width=3)
        self.crossLineHorShadow = pg.InfiniteLine(angle=0, movable=False, pen=self.crossShadowPen)
        self.crossLineVerShadow = pg.InfiniteLine(angle=90, movable=False, pen=self.crossShadowPen)
        self.crossLineHorizontal = pg.InfiniteLine(angle=0, movable=False, pen=self.crossPen)
        self.crossLineVertical = pg.InfiniteLine(angle=90, movable=False, pen=self.crossPen)

        self.imagePlotItem.addItem(self.crossLineVerShadow, ignoreBounds=True)
        self.imagePlotItem.addItem(self.crossLineHorShadow, ignoreBounds=True)
        self.imagePlotItem.addItem(self.crossLineVertical, ignoreBounds=True)
        self.imagePlotItem.addItem(self.crossLineHorizontal, ignoreBounds=True)

        self.probeLabel = pg.LabelItem('', justify='left')

        # Layout

        # Hiding the horCrossPlotItem and horCrossPlotItem will still leave some space in the
        # grid layout. We therefore remove them from the layout instead. We need to know if they
        # are already added.
        self.horPlotAdded = False
        self.verPlotAdded = False

        self.graphicsLayoutWidget = pg.GraphicsLayoutWidget()
        self.contentsLayout.addWidget(self.graphicsLayoutWidget)

        self.graphicsLayoutWidget.addItem(self.titleLabel, ROW_TITLE, COL_TITLE, colspan=3)
        self.graphicsLayoutWidget.addItem(self.colorLegendItem, ROW_COLOR, COL_COLOR, rowspan=2)
        self.graphicsLayoutWidget.addItem(self.imagePlotItem, ROW_IMAGE, COL_IMAGE)
        self.graphicsLayoutWidget.addItem(self.probeLabel, ROW_PROBE, COL_PROBE, colspan=3)

        gridLayout = self.graphicsLayoutWidget.ci.layout # A QGraphicsGridLayout
        gridLayout.setHorizontalSpacing(10)
        gridLayout.setVerticalSpacing(10)
        #gridLayout.setRowSpacing(ROW_PROBE, 40)

        gridLayout.setRowStretchFactor(ROW_HOR_LINE, 1)
        gridLayout.setRowStretchFactor(ROW_IMAGE, 2)
        gridLayout.setColumnStretchFactor(COL_IMAGE, 2)
        gridLayout.setColumnStretchFactor(COL_VER_LINE, 1)

        # Configuration tree
        self._config = PgImagePlot2dCti(pgImagePlot2d=self, nodeName='2D image plot')

        # Connect signals
        # Based mouseMoved on crosshair.py from the PyQtGraph examples directory.
        # I did not use the SignalProxy because I did not see any difference.
        self.imagePlotItem.scene().sigMouseMoved.connect(self.mouseMoved)


    def finalize(self):
        """ Is called before destruction. Can be used to clean-up resources.
        """
        logger.debug("Finalizing: {}".format(self))
        self.colorLegendItem.finalize()
        self.imagePlotItem.scene().sigMouseMoved.disconnect(self.mouseMoved)
        self.imagePlotItem.close()
        self.graphicsLayoutWidget.close()


    @classmethod
    def axesNames(cls):
        """ The names of the axes that this inspector visualizes.
            See the parent class documentation for a more detailed explanation.
        """
        return tuple(['Y', 'X'])


    def _clearContents(self):
        """ Clears the contents when no valid data is available
        """
        logger.debug("Clearing inspector contents")
        self.slicedArray = None
        self.titleLabel.setText('')

        # Don't clear the imagePlotItem, the imageItem is only added in the constructor.
        self.imageItem.clear()
        if hasattr(self.imageItem, '_wasIntegerData'):
            del self.imageItem._wasIntegerData

        # Unfortunately PyQtGraph doesn't emit this signal when the image is cleared.
        self.imageItem.sigImageChanged.emit()

        self.imagePlotItem.setLabel('left', '')
        self.imagePlotItem.setLabel('bottom', '')
        self.colorLegendItem.setLabel('')
        # Set the histogram range and levels to finite values to prevent futher errors if this
        # function was called after an exception in self.drawContents
        #self.histLutItem.setHistogramRange(0, 100)
        #self.histLutItem.setLevels(0, 100)

        self.crossPlotRow, self.crossPlotCol = None, None

        self.probeLabel.setText('')
        self.crossLineHorizontal.setVisible(False)
        self.crossLineVertical.setVisible(False)
        self.crossLineHorShadow.setVisible(False)
        self.crossLineVerShadow.setVisible(False)

        self.horCrossPlotItem.clear()
        self.verCrossPlotItem.clear()

        # Hide the complete widget. # TODO: do we still need the lines above?
        self.graphicsLayoutWidget.hide()


    def _drawContents(self, reason=None, initiator=None):
        """ Draws the plot contents from the sliced array of the collected repo tree item.

            The reason parameter is used to determine if the axes will be reset (the initiator
            parameter is ignored). See AbstractInspector.updateContents for their description.
        """

        # If auto-reset is true, reset config complete or partially, depending on the mode.
        if self._resetRequired(reason, initiator):
            self.resetConfig()

        self.crossPlotRow = None # reset because the sliced array shape may change
        self.crossPlotCol = None # idem dito

        gridLayout = self.graphicsLayoutWidget.ci.layout # A QGraphicsGridLayout

        if self.config.horCrossPlotCti.configValue:
            gridLayout.setRowStretchFactor(ROW_HOR_LINE, 1)
            if not self.horPlotAdded:
                self.graphicsLayoutWidget.addItem(self.horCrossPlotItem, ROW_HOR_LINE, COL_HOR_LINE)
                self.horPlotAdded = True
                gridLayout.activate()
        else:
            gridLayout.setRowStretchFactor(ROW_HOR_LINE, 0)
            if self.horPlotAdded:
                self.graphicsLayoutWidget.removeItem(self.horCrossPlotItem)
                self.horPlotAdded = False
                gridLayout.activate()

        if self.config.verCrossPlotCti.configValue:
            gridLayout.setColumnStretchFactor(COL_VER_LINE, 1)
            if not self.verPlotAdded:
                self.graphicsLayoutWidget.addItem(self.verCrossPlotItem, ROW_VER_LINE, COL_VER_LINE)
                self.verPlotAdded = True
                gridLayout.activate()
        else:
            gridLayout.setColumnStretchFactor(COL_VER_LINE, 0)
            if self.verPlotAdded:
                self.graphicsLayoutWidget.removeItem(self.verCrossPlotItem)
                self.verPlotAdded = False
                gridLayout.activate()

        slicedArray = self.collector.getSlicedArray()
        if slicedArray is None:
            self._clearContents()
            raise InvalidDataError()  # Don't show message, to common.
        elif not array_has_real_numbers(slicedArray.data):
            self._clearContents()
            raise InvalidDataError(
                "Selected item contains {} data.".format(array_kind_label(slicedArray.data)))
        else:
            self.slicedArray = slicedArray

        # -- Valid plot data from here on --

        self.graphicsLayoutWidget.show()

        if self.config.crossPlotGroupCti.checkState != Qt.Unchecked:
            tempPlotDataItem = self.config.crossPenCti.createPlotDataItem()
            if tempPlotDataItem.opts['pen'] is None and tempPlotDataItem.opts['symbol'] is None:
                self.sigShowMessage.emit(
                    "The cross-hair pen 'line' and 'symbol' config options are both unchecked!")

        numElem = np.prod(self.slicedArray.data.shape)
        if numElem == 0:
            self.sigShowMessage.emit("Current slice is empty.")  # Not expected to happen.
        elif numElem == 1:
            self.sigShowMessage.emit("Current slice contains only a single data point.")

        # PyQtGraph doesn't handle masked arrays so we convert the masked values to Nans. Missing
        # data values are replaced by NaNs. The PyQtGraph image plot shows this as the color at the
        # lowest end of the color scale. Unfortunately we cannot choose a missing-value color, but
        # at least the Nans do not influence for the histogram and color range.
        # We don't update self.slicedArray here because the data probe should still be able to
        # print the actual value.
        imageArray = replaceMaskedValueWithFloat(self.slicedArray.data, self.slicedArray.mask,
                                                 np.nan, copyOnReplace=True)

        # Replace infinite value with Nans because PyQtGraph fails on them. Note that the CTIs of
        # the cross plots (e.g. horCrossPlotRangeCti) are still connected to self.slicedArray, so
        # if the cross section consists of only infs, they may not able to update the autorange.
        # A warning is issued in that case.
        # We don't update self.slicedArray here because the data probe should still be able to
        # print the actual value.
        imageArray = replaceMaskedValueWithFloat(imageArray, np.isinf(self.slicedArray.data),
                                                 np.nan, copyOnReplace=True)

        # PyQtGraph uses the following dimension order: T, X, Y, Color.
        # We need to transpose the slicedArray ourselves because axes = {'x':1, 'y':0}
        # doesn't seem to do anything.
        imageArray = imageArray.transpose()

        # Set the _wasIntegerData to True if the original data type was a signed or unsigned. This
        # allows the ArgosColorLegendItem to make histogram bins as if it were an integer
        self.imageItem._wasIntegerData = self.slicedArray.data.dtype.kind in 'ui'
        self.imageItem.setAutoDownsample(self.config.autoDownSampleCti.configValue)
        self.imageItem.setImage(imageArray, autoLevels=False)  # Do after _wasIntegerData is set!

        self.imagePlotItem.setRectangleZoomOn(self.config.zoomModeCti.configValue)

        # Always use pan mode in the cross plots. Rectangle zoom is akward there and it's nice to
        # still be able to pan.
        #self.horCrossPlotItem.setRectangleZoomOn(self.config.zoomModeCti.configValue)
        #self.verCrossPlotItem.setRectangleZoomOn(self.config.zoomModeCti.configValue)

        self.horCrossPlotItem.invertX(self.config.xFlippedCti.configValue)
        self.verCrossPlotItem.invertY(self.config.yFlippedCti.configValue)

        self.probeLabel.setVisible(self.config.probeCti.configValue)

        self.titleLabel.setText(self.configValue('title').format(**self.collector.rtiInfo))

        # self.config.logBranch()
        self.config.updateTarget()


    @QtSlot(object)
    def mouseMoved(self, viewPos):
        """ Updates the probe text with the values under the cursor.
            Draws a vertical line and a symbol at the position of the probe.
        """
        try:
            check_class(viewPos, QtCore.QPointF)
            show_data_point = False # shows the data point as a circle in the cross hair plots
            self.crossPlotRow, self.crossPlotCol = None, None

            self.probeLabel.setText("<span style='color: #808080'>No data at cursor</span>")
            self.crossLineHorizontal.setVisible(False)
            self.crossLineVertical.setVisible(False)
            self.crossLineHorShadow.setVisible(False)
            self.crossLineVerShadow.setVisible(False)

            self.horCrossPlotItem.clear()
            self.verCrossPlotItem.clear()

            if self.slicedArray is not None and self.viewBox.sceneBoundingRect().contains(viewPos):

                # Calculate the row and column at the cursor.
                scenePos = self.viewBox.mapSceneToView(viewPos)
                row, col = round(scenePos.y()), round(scenePos.x())
                row, col = int(row), int(col) # Needed in Python 2
                nRows, nCols = self.slicedArray.shape

                if (0 <= row < nRows) and (0 <= col < nCols):
                    self.viewBox.setCursor(Qt.CrossCursor)

                    self.crossPlotRow, self.crossPlotCol = row, col
                    index = tuple([row, col])
                    valueStr = to_string(self.slicedArray.data[index],
                                         masked=self.slicedArray.maskAt(index),
                                         maskFormat='&lt;masked&gt;')

                    if self.config.probeCti.configValue:
                        txt = "({}, {}) = ({:d}, {:d}) {} {} = {}".format(
                            self.collector.rtiInfo['x-dim'], self.collector.rtiInfo['y-dim'],
                            col, row, RIGHT_ARROW, self.collector.rtiInfo['name'], valueStr)
                        self.probeLabel.setText(txt)

                    # Show cross section at the cursor pos in the line plots
                    if self.config.horCrossPlotCti.configValue:
                        self.crossLineHorShadow.setVisible(True)
                        self.crossLineHorizontal.setVisible(True)
                        self.crossLineHorShadow.setPos(row)
                        self.crossLineHorizontal.setPos(row)

                        # Line plot of cross section row.
                        # First determine which points are connected or separated by masks/nans.
                        rowData = self.slicedArray.data[row, :]
                        connected = np.isfinite(rowData)
                        if is_an_array(self.slicedArray.mask):
                            connected = np.logical_and(connected, ~self.slicedArray.mask[row, :])
                        else:
                            connected = (np.zeros_like(rowData)
                                         if self.slicedArray.mask else connected)

                        # Replace mask by Nans. Only doing when not showing lines to hack around PyQtGraph issue 1057
                        # See comment in PgLinePlot1d._drawContents for a more detailed explanation
                        # TODO: reuse imageItem data array when this hack is no longer necessary
                        if not self.config.crossPenCti.lineCti.configValue:
                            rowData = replaceMaskedValueWithFloat(rowData, np.logical_not(connected),
                                                                  np.nan, copyOnReplace=True)

                        # Replace infinite value with nans because PyQtGraph can't handle them
                        rowData = replaceMaskedValueWithFloat(rowData, np.isinf(rowData),
                                                              np.nan, copyOnReplace=True)

                        horPlotDataItem = self.config.crossPenCti.createPlotDataItem()
                        # TODO: try to use connect='finite' when the hack above is no longer necessary. In that case
                        # test with array_masked test data
                        horPlotDataItem.setData(rowData, connect=connected)
                        self.horCrossPlotItem.addItem(horPlotDataItem)


                        # Vertical line in hor-cross plot
                        crossLineShadow90 = pg.InfiniteLine(angle=90, movable=False,
                                                            pen=self.crossShadowPen)
                        crossLineShadow90.setPos(col)
                        self.horCrossPlotItem.addItem(crossLineShadow90, ignoreBounds=True)
                        crossLine90 = pg.InfiniteLine(angle=90, movable=False, pen=self.crossPen)
                        crossLine90.setPos(col)
                        self.horCrossPlotItem.addItem(crossLine90, ignoreBounds=True)

                        if show_data_point:
                            crossPoint90 = pg.PlotDataItem(symbolPen=self.crossPen)
                            crossPoint90.setSymbolBrush(QtGui.QBrush(
                                    self.config.crossPenCti.penColor))
                            crossPoint90.setSymbolSize(10)
                            crossPoint90.setData((col,), (rowData[col],))
                            self.horCrossPlotItem.addItem(crossPoint90, ignoreBounds=True)

                        self.config.horCrossPlotRangeCti.updateTarget() # update auto range
                        del rowData # defensive programming

                    if self.config.verCrossPlotCti.configValue:
                        self.crossLineVerShadow.setVisible(True)
                        self.crossLineVertical.setVisible(True)
                        self.crossLineVerShadow.setPos(col)
                        self.crossLineVertical.setPos(col)

                        # Line plot of cross section row.
                        # First determine which points are connected or separated by masks/nans.
                        colData = self.slicedArray.data[:, col]
                        connected = np.isfinite(colData)
                        if is_an_array(self.slicedArray.mask):
                            connected = np.logical_and(connected, ~self.slicedArray.mask[:, col])
                        else:
                            connected = (np.zeros_like(colData)
                                         if self.slicedArray.mask else connected)

                        # Replace mask by Nans. Only doing when not showing lines to hack around PyQtGraph issue 1057
                        # See comment in PgLinePlot1d._drawContents for a more detailed explanation
                        if not self.config.crossPenCti.lineCti.configValue:
                            colData = replaceMaskedValueWithFloat(colData, np.logical_not(connected),
                                                                  np.nan, copyOnReplace=True)

                        # Replace infinite value with nans because PyQtGraph can't handle them
                        colData = replaceMaskedValueWithFloat(colData, np.isinf(colData),
                                                              np.nan, copyOnReplace=True)

                        verPlotDataItem = self.config.crossPenCti.createPlotDataItem()
                        verPlotDataItem.setData(colData, np.arange(nRows), connect=connected)
                        self.verCrossPlotItem.addItem(verPlotDataItem)

                        # Horizontal line in ver-cross plot
                        crossLineShadow0 = pg.InfiniteLine(angle=0, movable=False,
                                                           pen=self.crossShadowPen)
                        crossLineShadow0.setPos(row)
                        self.verCrossPlotItem.addItem(crossLineShadow0, ignoreBounds=True)
                        crossLine0 = pg.InfiniteLine(angle=0, movable=False, pen=self.crossPen)
                        crossLine0.setPos(row)
                        self.verCrossPlotItem.addItem(crossLine0, ignoreBounds=True)

                        if show_data_point:
                            crossPoint0 = pg.PlotDataItem(symbolPen=self.crossPen)
                            crossPoint0.setSymbolBrush(QtGui.QBrush(self.config.crossPenCti.penColor))
                            crossPoint0.setSymbolSize(10)
                            crossPoint0.setData((colData[row],), (row,))
                            self.verCrossPlotItem.addItem(crossPoint0, ignoreBounds=True)

                        self.config.verCrossPlotRangeCti.updateTarget() # update auto range
                        del colData # defensive programming

        except Exception as ex:
            # In contrast to _drawContents, this function is a slot and thus must not throw
            # exceptions. The exception is logged. Perhaps we should clear the cross plots, but
            # this could, in turn, raise exceptions.
            if DEBUGGING:
                raise
            else:
                logger.exception(ex)




""" Classes to gather information about modules for the About page

    These classes may be slightly over-engineered but they serve also as an experimentation ground
    for the plug-in mechanism. We therefore define both an interface (AbstractModuleInfo) and a
    read-only class that can be used as a base for descendants, just to see if this is practical.
"""
from __future__ import print_function

import logging
import re
import sys

from abc import ABCMeta, abstractproperty

from argos.info import DEBUGGING

logger = logging.getLogger(__name__)

NOT_IMPLEMENTED_ERR_MSG = "You must override this abstract property/method."


def versionStrToTuple(versionStr):
    """ Converts a version string to tuple

        Version string has format <major>.<minor>.<patch><postfix> where postfix can be empty.
            E.g. versionStrToTuple('0.3.1rc1') returns (0, 3, 1, 'rc1')
    """
    matchObj = re.match(r"(\d+)\.(\d+)\.(\d+)(\w*)", versionStr)

    major = int(matchObj.group(1))
    minor = int(matchObj.group(2))
    patch = int(matchObj.group(3))
    postfix = matchObj.group(4)

    return (major, minor, patch, postfix)


class AbstractModuleInfo(object):
    """ Interface for the ModuleInfo classes
    """
    __metaclass__ = ABCMeta

    @abstractproperty
    def name(self):
        raise NotImplementedError(NOT_IMPLEMENTED_ERR_MSG)

    @abstractproperty
    def module(self):
        raise NotImplementedError(NOT_IMPLEMENTED_ERR_MSG)

    @abstractproperty
    def version(self):
        raise NotImplementedError(NOT_IMPLEMENTED_ERR_MSG)

    @abstractproperty
    def verboseVersion(self):
        raise NotImplementedError(NOT_IMPLEMENTED_ERR_MSG)

    @abstractproperty
    def packagePath(self):
        raise NotImplementedError(NOT_IMPLEMENTED_ERR_MSG)


class ReadOnlyModuleInfo(AbstractModuleInfo):
    """ Module information that can only set with the constructor.
        Most useful as a base class for descendants.
    """
    def __init__(self,
                 name='',
                 module=None,
                 version='',
                 verboseVersion='',
                 packagePath=''):

        self._name = name
        self._module = module
        self._version = version
        self._verboseVersion = verboseVersion
        self._packagePath = packagePath

    @property
    def name(self):
        return self._name

    @property
    def module(self):
        return self._module

    @property
    def version(self):
        return self._version

    @property
    def verboseVersion(self):
        return self._verboseVersion

    @property
    def packagePath(self):
        return self._packagePath



# It is impossible to derive from ReadOnlyModuleInfo and only define the setters. Python
# will give an error because it would violate the contract that the properties cannot be changed.
# Therefore we make a simple class, which uses attributes, to use for debugging.
class DuckTypingModuleInfo(object):
    """ Module information that can be written to after contstruction.
        Most useful for debugging purposes.

        Does not inherit from the interface but uses the duck-typing principle
    """
    def __init__(self, name,
                 module=None,
                 version='',
                 verboseVersion='',
                 packagePath=''):

        self.name = name
        self.module = module
        self.version = version
        self.verboseVersion = verboseVersion
        self.packagePath = packagePath



class PythonModuleInfo(ReadOnlyModuleInfo):

    def __init__(self):

        super(PythonModuleInfo, self).__init__('Python', module=None)
        self._version = "{0.major}.{0.minor}.{0.micro}".format(sys.version_info)
        self._verboseVersion = sys.version.replace('\n', '')


class ImportedModuleInfo(ReadOnlyModuleInfo):
    """ Tries to import a module by name and retrieve information from it.
    """
    def __init__(self, name,
                 module=None,
                 verboseVersion=None,
                 versionAttribute='__version__',
                 pathAttribute='__path__'):

        super(ImportedModuleInfo, self).__init__(name, module = module)

        self._versionAttribute = versionAttribute
        self._pathAttribute = pathAttribute

        if module is None:
            self.tryImportModule(name)

        if verboseVersion is None:
            self._verboseVersion = self._version


    def tryImportModule(self, name):
        """ Imports the module and sets version information
            If the module cannot be imported, the version is set to empty values.
        """
        self._name = name
        try:
            import importlib
            self._module = importlib.import_module(name)
        except ImportError:
            self._module = None
            self._version = ''
            self._packagePath = ''
        else:
            if self._versionAttribute:
                self._version = getattr(self._module, self._versionAttribute, '???')
            if self._pathAttribute:
                self._packagePath = getattr(self._module, self._pathAttribute, '???')


#################
# Special cases #
#################


class ArgosModuleInfo(ImportedModuleInfo):

    def __init__(self):

        super(ArgosModuleInfo, self).__init__('argos')
        if self.module:
            self._verboseVersion = ('{}{}'
                                    .format(self._version, ' (debugging-mode)' if DEBUGGING else ''))


class H5pyModuleInfo(ImportedModuleInfo):

    def __init__(self):

        super(H5pyModuleInfo, self).__init__('h5py', versionAttribute=None)
        if self.module:
            self._version = self.module.version.version
            self._verboseVersion = ('{} (libhdf5: {})'
                                    .format(self._version, self.module.version.hdf5_version))


class NetCDF4ModuleInfo(ImportedModuleInfo):

    def __init__(self):

        super(NetCDF4ModuleInfo, self).__init__('netCDF4', pathAttribute=None)
        if self.module:
            self._verboseVersion = ('{} (libncdf4: {}, libhdf5: {})'
                                    .format(self._version,
                                            self.module.__netcdf4libversion__,
                                            self.module.__hdf5libversion__))


class PillowInfo(ImportedModuleInfo):

    def __init__(self):

        super(PillowInfo, self).__init__('PIL')
        self._name = 'pillow (PIL)'


class QtModuleInfo(ImportedModuleInfo):

    def __init__(self):
        import argos.qt.bindings as qtbind
        super(QtModuleInfo, self).__init__(name=qtbind.QT_API_NAME, module=qtbind,
                                           versionAttribute='PYQT_VERSION', pathAttribute=None)

        self._verboseVersion = qtbind.PYQT_VERSION + " ("
        # if qtmisc.QT_API_NAME:
        #     self._verboseVersion += "api {}, qtpy: {}, ".format(qtmisc.QT_API)
        self._verboseVersion += "Qt: {})".format(qtbind.QT_VERSION)




import logging

from argos.external import six
from argos.utils.cls import check_class, check_is_a_string

logger = logging.getLogger(__name__)

class BaseTreeItem(object):
    """ Base class for storing item data in a tree form. Each tree item represents a row
        in the BaseTreeModel (QAbstractItemModel).

        The tree items have no notion of which field is stored in which column. This is implemented
        in BaseTreeModel._itemValueForColumn
    """
    def __init__(self, nodeName):
        """ Constructor

            :param nodeName: short name describing this node. Is used to construct the nodePath.
                Currently we don't check for uniqueness in the children but this may change.
                The nodeName may not contain slashes (/).
        """
        check_class(nodeName, six.string_types, allow_none=False)
        assert nodeName, "nodeName may not be empty"
        assert '/' not in nodeName, "nodeName may not contain slashes"
        self._nodeName = str(nodeName)
        self._parentItem = None
        self._model = None
        self._childItems = [] # the fetched children
        self._nodePath = self._constructNodePath()

    def finalize(self):
        """ Can be used to cleanup resources. Should be called explicitly.
            Finalizes its children before closing itself
        """
        for child in self.childItems:
            child.finalize()

    def __str__(self):
        return "<{}: {}>".format(type(self).__name__, self.nodePath)

    def __repr__(self):
        return ("<{}: {!r}, children:[{}]>".
                format(type(self).__name__, self.nodePath,
                       ', '.join([repr(child) for child in self.childItems])))

    @property
    def model(self):
        """ Returns the ConfigTreeModel this item belongs to.
            If the model is None (not set), it will use and cache the parent's model.
            Therefore make sure that an ancestor node has a reference to the model! Typically by
            setting the model property of the invisible root item in the model constructor.
        """
        if self._model is None and self.parentItem is not None:
            self._model = self.parentItem.model
        return self._model

    @model.setter
    def model(self, value):
        """ Sets ConfigTreeModel this item belongs to.
        """
        self._model = value


    @property
    def decoration(self):
        """ An optional decoration (e.g. icon).
            The default implementation returns None (no decoration).
        """
        return None

    @property
    def font(self):
        """ Returns a font for displaying this item's text in the tree.
            The default implementation returns None (i.e. uses default font).
        """
        return None

    @property
    def backgroundBrush(self):
        """ Returns a brush for drawing the background role in the tree.
            The default implementation returns None (i.e. uses default brush).
        """
        return None

    @property
    def foregroundBrush(self):
        """ Returns a brush for drawing the foreground role in the tree.
            The default implementation returns None (i.e. uses default brush).
        """
        return None

    @property
    def sizeHint(self):
        """ Returns a size hint for displaying the items in the tree
            The default implementation returns None (i.e. no hint).
            Should return a QSize object or None
        """
        return None

    @property
    def nodeName(self):
        """ The node name. Is used to construct the nodePath"""
        return self._nodeName

    @nodeName.setter
    def nodeName(self, nodeName):
        """ The node name. Is used to construct the nodePath"""
        assert '/' not in nodeName, "nodeName may not contain slashes"
        self._nodeName = nodeName
        self._recursiveSetNodePath(self._constructNodePath())

    def _constructNodePath(self):
        """ Recursively prepends the parents nodeName to the path until the root node is reached."""
        if self.parentItem is None:
            return '' # invisible root node; is not included in the path
        else:
            return self.parentItem.nodePath + '/' + self.nodeName

    @property
    def nodePath(self):
        """ The sequence of nodeNames from the root to this node. Separated by slashes."""
        return self._nodePath

    def _recursiveSetNodePath(self, nodePath):
        """ Sets the nodePath property and updates it for all children.
        """
        self._nodePath = nodePath
        for childItem in self.childItems:
            childItem._recursiveSetNodePath(nodePath + '/' + childItem.nodeName)

    @property
    def parentItem(self):
        """ The parent item """
        return self._parentItem

    @parentItem.setter
    def parentItem(self, value):
        """ The parent item """
        self._parentItem = value
        self._recursiveSetNodePath(self._constructNodePath())

    @property
    def childItems(self):
        """ List of child items """
        #logger.debug("childItems {!r}".format(self))
        return self._childItems

    def hasChildren(self):
        """ Returns True if the item has children.

            If it has children the corresponding node in the tree can be expanded.
        """
        return len(self.childItems) > 0

    def nChildren(self): # TODO: numChildren
        """ Returns the number of children
        """
        return len(self.childItems)

    def child(self, row):
        """ Gets the child given its row number
        """
        return self.childItems[row]


    def childByNodeName(self, nodeName):
        """ Gets first (direct) child that has the nodeName.
        """
        assert '/' not in nodeName, "nodeName can not contain slashes"
        for child in self.childItems:
            if child.nodeName == nodeName:
                return child

        raise IndexError("No child item found having nodeName: {}".format(nodeName))


    def findByNodePath(self, nodePath):
        """ Recursively searches for the child having the nodePath. Starts at self.
        """
        def _auxGetByPath(parts, item):
            "Aux function that does the actual recursive search"
            #logger.debug("_auxGetByPath item={}, parts={}".format(item, parts))

            if len(parts) == 0:
                return item

            head, tail = parts[0], parts[1:]
            if head == '':
                # Two consecutive slashes. Just go one level deeper.
                return _auxGetByPath(tail, item)
            else:
                childItem = item.childByNodeName(head)
                return _auxGetByPath(tail, childItem)

        # The actual body of findByNodePath starts here

        check_is_a_string(nodePath)
        assert not nodePath.startswith('/'), "nodePath may not start with a slash"

        if not nodePath:
            raise IndexError("Item not found: {!r}".format(nodePath))

        return _auxGetByPath(nodePath.split('/'), self)


    def childNumber(self):
        """ Gets the index (nr) of this node in its parent's list of children.
        """
        # This is O(n) in time. # TODO: store row number in the items?
        if self.parentItem is not None:
            return self.parentItem.childItems.index(self)
        return 0


    def insertChild(self, childItem, position=None):
        """ Inserts a child item to the current item.
            The childItem must not yet have a parent (it will be set by this function).

            IMPORTANT: this does not let the model know that items have been added.
            Use BaseTreeModel.insertItem instead.

            param childItem: a BaseTreeItem that will be added
            param position: integer position before which the item will be added.
                If position is None (default) the item will be appended at the end.

            Returns childItem so that calls may be chained.
        """
        if position is None:
            position = self.nChildren()

        assert childItem.parentItem is None, "childItem already has a parent: {}".format(childItem)
        assert childItem._model is None, "childItem already has a model: {}".format(childItem)

        childItem.parentItem = self
        childItem.model = self.model
        self.childItems.insert(position, childItem)
        return childItem


    def removeChild(self, position):
        """ Removes the child at the position 'position'
            Calls the child item finalize to close its resources before removing it.
        """
        assert 0 <= position <= len(self.childItems), \
            "position should be 0 < {} <= {}".format(position, len(self.childItems))

        self.childItems[position].finalize()
        self.childItems.pop(position)


    def removeAllChildren(self):
        """ Removes the all children of this node.
            Calls the child items finalize to close their resources before removing them.
        """
        for childItem in self.childItems:
            childItem.finalize()
        self._childItems = []


    def logBranch(self, indent=0, level=logging.DEBUG):
        """ Logs the item and all descendants, one line per child
        """
        if 0:
            print(indent * "    " + str(self))
        else:
            logger.log(level, indent * "    " + str(self))
        for childItems in self.childItems:
            childItems.logBranch(indent + 1, level=level)



class AbstractLazyLoadTreeItem(BaseTreeItem):
    """ Abstract base class for a tree item that can do lazy loading of children.
        Descendants should override the _fetchAllChildren
    """
    def __init__(self, nodeName=''):
        """ Constructor
        """
        super(AbstractLazyLoadTreeItem, self).__init__(nodeName=nodeName)
        self._canFetchChildren = True # children not yet fetched (successfully or unsuccessfully)


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return self._canFetchChildren or len(self.childItems) > 0


    def canFetchChildren(self):
        """ Returns True if children can be fetched, and False if they already have been fetched.
            Also returns False if they have been fetched and tried.
        """
        return self._canFetchChildren


    def fetchChildren(self):
        """ Fetches children.

            The actual work is done by _fetchAllChildren. Descendant classes should typically
            override that method instead of this one.
        """
        assert self._canFetchChildren, "canFetchChildren must be True"
        try:
            childItems = self._fetchAllChildren()
        finally:
            self._canFetchChildren = False # Set to True, even if tried and failed.

        return childItems


    def _fetchAllChildren(self):
        """ The function that actually fetches the children.

            The result must be a list of RepoTreeItems. Their parents must be None,
            as that attribute will be set by BaseTreeitem.insertItem()

            :rtype: list of BaseRti objects
        """
        raise NotImplementedError


    def removeAllChildren(self):
        """ Removes all children """
        try:
            super(AbstractLazyLoadTreeItem, self).removeAllChildren()
        finally:
            self._canFetchChildren = True



""" Contains ScientificDoubleSpinBox, a QDoubleSpinBox that can handle scientific notation.

    Gratefully copied and adapted from:
        https://www.jdreaver.com/posts/2014-07-28-scientific-notation-spin-box-pyside.html

"""
from __future__ import division

import re, logging
import numpy as np
from argos.qt import QtGui, QtWidgets

logger = logging.getLogger(__name__)

# Regular expression to find floats. Match groups are the whole string, the
# whole coefficient, the decimal part of the coefficient, and the exponent
# part.
REGEXP_FLOAT = re.compile(r'(([+-]?\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)')


def valid_float_string(string):
    match = REGEXP_FLOAT.search(string)
    return match.groups()[0] == string if match else False


class FloatValidator(QtGui.QValidator):

    def validate(self, string, position):
        if valid_float_string(string):
            return (QtGui.QValidator.Acceptable, string, position)
        if string == "" or string[position-1] in 'e.-+':
            return (QtGui.QValidator.Intermediate, string, position)
        return (QtGui.QValidator.Invalid , string, position)

    def fixup(self, text):
        match = REGEXP_FLOAT.search(text)
        return match.groups()[0] if match else ""



class ScientificDoubleSpinBox(QtWidgets.QDoubleSpinBox):
    """ A QDoubleSpinBox that can handle scientific notation.
    """
    def __init__(self,
                 precision=6,
                 largeStepFactor=10,
                 smallStepsPerLargeStep=10,
                 *args, **kwargs):
        """ Constructor.

            In contrast to the QDoubleSpinbox the (page)up/down let the value grow exponentially.
            That is, the spinbox value is multiplied by a small step factor when the up-arrow is
            pressed, and by a larger factor if page-up is pressed. A large step multiplies the value
            spinbox value with largeStepFactor (default 10). The smallStepsPerLargeStep does then
            specify how many up-arrow key presses are needed to increase to largeStepFactor.

            :param precision: The precision used in the scientific notation.
            :param largeStepFactor: default 10
            :param smallStepsPerLargeStep: default 10
        """
        self.precision = precision
        super(ScientificDoubleSpinBox, self).__init__(*args, **kwargs)
        self.setMinimum(np.finfo('d').min)
        self.setMaximum(np.finfo('d').max)
        self.validator = FloatValidator()
        self.setDecimals(323) # because of the limitations of the double type (see Qt docs).

        self._smallStepFactor = None # determined by _smallStepsPerLargeStep and _largeStepFactor
        self._smallStepsPerLargeStep = None
        self._largeStepFactor = largeStepFactor
        self.smallStepsPerLargeStep = smallStepsPerLargeStep


    def validate(self, text, position):
        result = self.validator.validate(text, position)
        #logger.debug("validate result: {}".format(result))
        return result


    def fixup(self, text):
        result = self.validator.fixup(text)
        #logger.debug("fixup: {}".format(result))
        return result


    def valueFromText(self, text):
        return float(text)


    def textFromValue(self, value):
        return "{:.{precission}g}".format(value, precission=self.precision)


    @property
    def largeStepFactor(self):
        """ The spinbox will be multiplied with this factor whenever page up is pressed.
        """
        return self._largeStepFactor


    @largeStepFactor.setter
    def largeStepFactor(self, largeStepFactor):
        """ The spinbox will be multiplied with this factor whenever page up is pressed.
        """
        self._largeStepFactor = largeStepFactor


    @property
    def smallStepFactor(self):
        """ The spinbox will be multiplied with this factor whenever page up is pressed.
            Read-only property. Setting is done via largeStepFactor and smallStepsPerLargeStep.
        """
        return self._smallStepFactor


    @property
    def smallStepsPerLargeStep(self):
        """ The number of small steps that go in a large one.

            The spinbox value is increased with a small step when the up-arrow is pressed, and by
            a large step if page-up is pressed. A large step increases the value increases the
            spinbox value with largeStepFactor (default 10). The smallStepsPerLargeStep does then
            specify how many up-arrow key presses are needed to increase to largeStepFactor.
        """
        return self._smallStepsPerLargeStep


    @smallStepsPerLargeStep.setter
    def smallStepsPerLargeStep(self, smallStepsPerLargeStep):
        """ Sets the number of small steps that go in a large one.

        """
        self._smallStepsPerLargeStep = smallStepsPerLargeStep
        self._smallStepFactor = np.power(self.largeStepFactor, 1.0 / smallStepsPerLargeStep)


    def stepBy(self, steps):
        """ Function that is called whenever the user triggers a step. The steps parameter
            indicates how many steps were taken, e.g. Pressing Qt::Key_Down will trigger a call to
            stepBy(-1), whereas pressing Qt::Key_Prior will trigger a call to stepBy(10).
        """
        oldValue = self.value()

        if oldValue == 0:
            newValue = steps
        elif steps == 1:
            newValue = self.value() * self.smallStepFactor
        elif steps == -1:
            newValue = self.value() / self.smallStepFactor
        elif steps == 10:
            newValue = self.value() * self.largeStepFactor
        elif steps == -10:
            newValue = self.value() / self.largeStepFactor
        else:
            raise ValueError("Invalid step size: {!r}, value={}".format(steps, oldValue))

        newValue = float(newValue)

        if newValue < self.minimum():
            newValue = self.minimum()

        if newValue > self.maximum():
            newValue = self.maximum()

        #logger.debug("stepBy {}: {} -> {}".format(steps, oldValue, newValue))
        try:
            self.setValue(newValue)
        except Exception:
            # TODO: does this ever happen? Better validation (e.g. catch underflows)
            logger.warning("Unable to set spinbox to: {!r}".format(newValue))
            self.setValue(oldValue)



if __name__ == "__main__":
    import sys

    logger = logging.getLogger(__name__)
    logging.basicConfig(level='DEBUG', stream=sys.stderr,
        format='%(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: %(message)s')

    def main():
        """ Small stand-alone test
        """
        app = QtWidgets.QApplication(sys.argv[1:])
        spinBox = ScientificDoubleSpinBox(precision = 9, largeStepFactor=2, smallStepsPerLargeStep=3)
        spinBox.selectAll()
        spinBox.setSingleStep(2)
        spinBox.raise_()
        spinBox.show()
        sys.exit(app.exec_())


    #######
    main()



# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Dialog window that allows users to configure the installed plugins.
"""
from __future__ import print_function

import copy
import logging
import os.path
import sys

from argos.info import icons_directory
from argos.qt import QtCore, QtGui, QtWidgets, Qt, QtSlot
from argos.qt.colorselect import ColorSelectWidget
from argos.qt.shortcutedit import ShortCutEditor
from argos.reg.basereg import BaseRegistryModel, BaseRegistry, RegType
from argos.reg.tabview import TableEditWidget
from argos.utils.cls import check_class
from argos.utils.misc import wrapHtmlColor
from argos.widgets.constants import MONO_FONT, FONT_SIZE, COLOR_ERROR
#from argos.widgets.constants import QCOLOR_REGULAR, QCOLOR_NOT_IMPORTED, QCOLOR_ERROR

logger = logging.getLogger(__name__)

# The main window inherits from a Qt class, therefore it has many
# ancestors public methods and attributes.
# pylint: disable=R0901, R0902, R0904, W0201



class PluginsDialog(QtWidgets.QDialog):
    """ Dialog window that allows users to configure the installed plugins.
    """

    def __init__(self, label, registry,  parent=None):
        """ Constructor
        """
        super(PluginsDialog, self).__init__(parent=parent)
        check_class(registry, BaseRegistry)

        self._label = label
        self._orgRegistry = registry
        self._registry = copy.deepcopy(registry)  # make copy so changes can be canceled
        self._tableModel = self._registry.createTableModel(parent=self)
        self.mapper = QtWidgets.QDataWidgetMapper(parent=self)
        self.mapper.setModel(self._tableModel)

        self.setWindowTitle("Argos {} Plugins".format(label))

        layout = QtWidgets.QVBoxLayout(self)

        self.verSplitter = QtWidgets.QSplitter(Qt.Vertical)
        #self.verSplitter.setCollapsible(1, False)
        self.verSplitter.setChildrenCollapsible(False)
        layout.addWidget(self.verSplitter)

        self.tableWidget = TableEditWidget(self._tableModel)
        self.verSplitter.addWidget(self.tableWidget)

        self._tableView = self.tableWidget.tableView
        self._tableView.installEventFilter(self)

        # Form
        self.horSplitter = QtWidgets.QSplitter(Qt.Horizontal)
        self.horSplitter.setChildrenCollapsible(False)

        self.verSplitter.addWidget(self.horSplitter)

        self.formWidget = QtWidgets.QWidget()
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setContentsMargins(0, 0, 0, 0)
        self.formWidget.setLayout(self.formLayout)
        self.horSplitter.addWidget(self.formWidget)

        self._editWidgets = []
        itemCls = registry.ITEM_CLASS
        assert len(itemCls.LABELS) == len(itemCls.TYPES), \
            "Regtype Mismatch: {} != {}".format(len(itemCls.LABELS), len(itemCls.TYPES))
        for col, (label, regType) in enumerate(zip(itemCls.LABELS, itemCls.TYPES)):

            if regType == RegType.String:
                editWidget = QtWidgets.QLineEdit()
                self.mapper.addMapping(editWidget, col)
            elif regType == RegType.ShortCut:
                editWidget = ShortCutEditor()
                self.mapper.addMapping(editWidget, col)
            elif regType == RegType.ColorStr:
                editWidget = ColorSelectWidget()
                self.mapper.addMapping(editWidget.lineEditor, col)
            else:
                raise AssertionError("Unexpected regType: {}".format(regType))
            editWidget.installEventFilter(self)
            self.formLayout.addRow(label, editWidget)
            self._editWidgets.append(editWidget)

        # Detail info widget
        font = QtGui.QFont()
        font.setFamily(MONO_FONT)
        font.setFixedPitch(True)
        font.setPointSize(FONT_SIZE)

        self.editor = QtWidgets.QTextEdit()
        self.editor.setReadOnly(True)
        #self.editor.setFocusPolicy(Qt.NoFocus) # Allow focus so that user can copy text from it.
        #self.editor.setFont(font)
        self.editor.setWordWrapMode(QtGui.QTextOption.WordWrap)
        self.editor.clear()
        self.horSplitter.addWidget(self.editor)

        self.horSplitter.setStretchFactor(0, 2)
        self.horSplitter.setStretchFactor(1, 3)
        self.verSplitter.setSizes([300, 150])

        # Reset/Cancel/Save Buttons

        self.saveButton = QtWidgets.QPushButton("Save")
        self.saveButton.clicked.connect(self.accept)

        self.cancelButton = QtWidgets.QPushButton("Cancel")
        self.cancelButton.clicked.connect(self.reject)

        self.resetButton = QtWidgets.QPushButton("Reset Table to Defaults...")
        self.resetButton.clicked.connect(self.resetToDefaults)
        self.resetButton.setIcon(QtGui.QIcon(os.path.join(icons_directory(), 'reset-l.svg')))

        # We use a button layout instead of a QButtonBox because there always will be a default
        # button (e.g. the Save button) that will light up, even if another widget has the focus.
        # From https://doc.qt.io/archives/qt-4.8/qdialogbuttonbox.html#details
        #   However, if there is no default button set and to preserve which button is the default
        #   button across platforms when using the QPushButton::autoDefault property, the first
        #   push button with the accept role is made the default button when the QDialogButtonBox
        #   is shown,

        self.buttonLayout = QtWidgets.QHBoxLayout()
        self.buttonLayout.addWidget(self.resetButton)
        self.buttonLayout.addStretch()
        if sys.platform == 'darwin':
            self.buttonLayout.addWidget(self.cancelButton)
            self.buttonLayout.addWidget(self.saveButton)
        else:
            self.buttonLayout.addWidget(self.saveButton)
            self.buttonLayout.addWidget(self.cancelButton)
        layout.addLayout(self.buttonLayout)

        # Connect signals and populate

        self.tableWidget.tableView.selectionModel().currentChanged.connect(self.currentItemChanged)
        self.tableWidget.tableView.model().sigItemChanged.connect(self._updateEditor)

        self.resize(QtCore.QSize(1100, 700))
        self.tableWidget.tableView.setFocus(Qt.NoFocusReason)

        self.tryImportAllPlugins()



    def eventFilter(self, targetWidget, event):
        """ Toggles the focus between table and edit form when the return key is pressed.
        """
        if event.type() == QtCore.QEvent.KeyPress and event.key() in (Qt.Key_Enter, Qt.Key_Return):

            if (event.modifiers() & Qt.ControlModifier):
                self.accept()
                return True

            if targetWidget in self._editWidgets:
                self._tableView.setFocus()
                return True

            elif targetWidget == self._tableView:
                curIdx = self._tableView.currentIndex()
                col = curIdx.column() if curIdx.isValid() else 0
                self._editWidgets[col].setFocus()
                return True

        # Give parent event filter chance to filter the event. Otherwise targetWidget will get it.
        return super(PluginsDialog, self).eventFilter(targetWidget, event)


    def accept(self):
        """ Saves registry.

            After saving the application may be in an inconsistent state. For instance, files
            may be opened with plugins that no longer exist. Therefore the caller must 'restart'
            the application if the changes were accepted.
        """
        logger.debug("Updating registry from dialog")

        # Copy contents from copy back to the original registry
        self._orgRegistry.clear()
        self._orgRegistry.unmarshall(self._registry.marshall())
        super(PluginsDialog, self).accept()


    def resetToDefaults(self):
        """ Resets the registry to it's default values
        """
        button = QtWidgets.QMessageBox.question(
            self, "Reset {}".format(self._label), "Reset plugins to the default set?")

        if button != QtWidgets.QMessageBox.Yes:
            logger.info("Resetting registry canceled.")
            return None

        logger.info("Resetting {} registry to its default.".format(self._label))
        self._tableModel.beginResetModel()
        try:
            self._registry.unmarshall(cfg={}) #  Empty config will cause the defaults to be used.
            self.tryImportAllPlugins()
        finally:
            self._tableModel.endResetModel()


    def tryImportAllPlugins(self):
        """ Tries to import all underlying plugin classes
        """
        logger.debug("Importing all plugins.")

        for regItem in self._registry.items:
            if not regItem.triedImport:
                self._tableModel.tryImportRegItem(regItem)

        logger.debug("Importing finished.")


    def getCurrentRegItem(self):
        """ Returns the item that is currently selected in the table.
            Can return None if there is no data in the table
        """
        return self.tableWidget.tableView.getCurrentItem()


    @QtSlot(QtCore.QModelIndex, QtCore.QModelIndex)
    def currentItemChanged(self, currentIndex=None, _previousIndex=None):
        """ Updates the description text widget when the user clicks on a selector in the table.
            The _currentIndex and _previousIndex parameters are ignored.
        """
        self.mapper.setCurrentModelIndex(currentIndex)
        regItem = self.getCurrentRegItem()
        self._updateEditor(regItem)


    def _updateEditor(self, regItem):
        """ Updates the editor with contents of the currently selected regItem
        """
        self.editor.clear()

        if regItem is None:
            return

        header = "<h2>{}</h2>".format(regItem.name)

        if regItem.successfullyImported is None:
            html = header + 'Plugin not yet imported!'
            self.editor.setHtml(wrapHtmlColor(html, COLOR_ERROR))
        elif regItem.successfullyImported is False:
            html = "{}Unable to import plugin.\n\nError: {}".format(header, regItem.exception)
            if hasattr(regItem.exception, "traceBackString"):
                html += "<pre>{}</pre>".format(regItem.exception.traceBackString)
            self.editor.setHtml(wrapHtmlColor(html, COLOR_ERROR))
        elif regItem.descriptionHtml:
            self.editor.setHtml(header.replace('\n', '<br>') + regItem.descriptionHtml)
        else:
            self.editor.setHtml(header + regItem.docString.replace('\n', '<br>'))




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Base class for inspectors
"""
import logging

from argos.info import DEBUGGING
from argos.qt import QtSignal, QtCore, QtWidgets
from argos.qt.togglecolumn import ToggleColumnTableWidget
from argos.repo.baserti import BaseRti
from argos.utils.cls import get_class_name, check_class
from argos.widgets.constants import DOCK_SPACING, DOCK_MARGIN, LEFT_DOCK_WIDTH
from argos.widgets.display import MessageDisplay
from argos.widgets.misc import BasePanel

logger = logging.getLogger(__name__)



class DetailBasePane(QtWidgets.QStackedWidget):
    """ Base class for plugins that show details of the current repository tree item.
        Serves as an interface but can also be instantiated for debugging purposes.
        A detail pane is a stacked widget; it has a contents page and and error page.
    """
    _label = "Details"

    ERROR_PAGE_IDX = 0
    CONTENTS_PAGE_IDX = 1

    sigUpdated = QtSignal(bool)  # Parameter indicates sucess

    def __init__(self, repoTreeView, parent=None):
        """ Constructor takes a reference to the repository tree view it monitors
        """
        super(DetailBasePane, self).__init__(parent)

        self._isConnected = False
        self._repoTreeView = repoTreeView

        self.errorWidget = MessageDisplay()
        self.addWidget(self.errorWidget)

        self.contentsWidget = BasePanel()
        self.addWidget(self.contentsWidget)

        self.contentsLayout = QtWidgets.QBoxLayout(QtWidgets.QBoxLayout.TopToBottom)
        self.contentsLayout.setSpacing(DOCK_SPACING)
        self.contentsLayout.setContentsMargins(DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN)
        self.contentsWidget.setLayout(self.contentsLayout)

        self.setCurrentIndex(self.CONTENTS_PAGE_IDX)


    @classmethod
    def classLabel(cls):
        """ Returns a short string that describes this class. For use in menus, headers, etc.
        """
        return cls._label


    @property
    def isConnected(self):
        "Returns True if this pane is connected to the currentChanged signal of the repoTreeView"
        return self._isConnected


    def sizeHint(self):
        """ The recommended size for the widget."""
        return QtCore.QSize(LEFT_DOCK_WIDTH, 250)


    def repoItemChanged(self, rti):
        """ Updates the content when the current repo tree item changes.
            The rti parameter can be None when no RTI is selected in the repository tree.
        """
        check_class(rti, (BaseRti, int), allow_none=True)
        assert type(rti) != int, "rti: {}".format(rti)
        try:
            self._drawContents(rti)
            self.setCurrentIndex(self.CONTENTS_PAGE_IDX)
            if rti is not None:
                self.sigUpdated.emit(True)
        except Exception as ex:
            if DEBUGGING:
                raise
            logger.exception(ex)
            self.errorWidget.setError(msg=str(ex), title=get_class_name(ex))
            self.setCurrentIndex(self.ERROR_PAGE_IDX)
            if rti is not None:
                self.sigUpdated.emit(False)


    def _drawContents(self, currentRti=None):
        """ Draws the contents for the current RTI. Descendants should override this.
            Descendants should draw 'empty' contents if currentRti is None. No need to
            handle exceptions though, these are handled by the called (currentChanged).
        """
        pass


    def marshall(self):
        """ Returns a dictionary to save in the persistent settings
        """
        raise NotImplementedError("Not implemented. Please override")


    def unmarshall(self, cfg):
        """ Initializes itself from a config dict form the persistent settings.
        """
        raise NotImplementedError("Not implemented. Please override")


class DetailTablePane(DetailBasePane):
    """ Base class for inspectors that consist of a single QTableWidget
    """
    _label = "Details Table"

    HEADERS = [] # Children should override this

    def __init__(self, repoTreeView, parent=None):
        super(DetailTablePane, self).__init__(repoTreeView, parent=parent)

        self.table = ToggleColumnTableWidget()
        self.contentsLayout.addWidget(self.table)

        self.table.setWordWrap(False)
        #self.table.setTextElideMode(QtCore.Qt.ElideNone)
        self.table.setColumnCount(len(self.HEADERS))
        self.table.setHorizontalHeaderLabels(self.HEADERS)
        self.table.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.table.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.table.verticalHeader().hide()
        self.table.setAlternatingRowColors(True)
        self.table.setShowGrid(False)

        self.table.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)

        tableHeader = self.table.horizontalHeader()
        tableHeader.setSectionResizeMode(QtWidgets.QHeaderView.Interactive) # don't set to stretch
        tableHeader.setStretchLastSection(True)


    def marshall(self):
        """ Returns a dictionary to save in the persistent settings
        """
        cfg = dict(tableHeaders=self.table.marshall())
        return cfg


    def unmarshall(self, cfg):
        """ Initializes itself from a config dict form the persistent settings.
        """
        self.table.unmarshall(cfg.get('tableHeaders'))




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Inspector that only shows an (error) message
"""
import logging

from argos.config.groupcti import MainGroupCti

from argos.inspector.abstract import AbstractInspector
from argos.utils.cls import check_is_a_string


logger = logging.getLogger(__name__)


class ErrorMsgInspector(AbstractInspector):
    """ Inspector that shows an error message

        :param str msg: the string that will be
    """
    def __init__(self, collector, msg, parent=None):

        super(ErrorMsgInspector, self).__init__(collector, parent=parent)

        check_is_a_string(msg)
        self.msg = msg

        self._config = self._createConfig()

        self.setCurrentIndex(self.ERROR_PAGE_IDX)


    @classmethod
    def axesNames(cls):
        """ The names of the axes that this inspector visualizes.
            See the parent class documentation for a more detailed explanation.
        """
        return tuple()


    def _createConfig(self):
        """ Creates a config tree item (CTI) hierarchy containing default children.
        """
        rootItem = MainGroupCti('message inspector')
        return rootItem


    def updateContents(self, reason=None, initiator=None):
        """ Override updateContents. Shows the error message
        """
        self.setCurrentIndex(self.ERROR_PAGE_IDX)
        self._showError(msg=self.msg, title="Error")




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Module that contains ArgosPgPlotItem.
"""
from __future__ import division, print_function

import logging
import pyqtgraph as pg

from argos.info import DEBUGGING
from argos.qt import Qt, QtCore, QtWidgets, QtSignal

logger = logging.getLogger(__name__)

from pyqtgraph.graphicsItems.PlotItem import PlotItem


X_AXIS = pg.ViewBox.XAxis
Y_AXIS = pg.ViewBox.YAxis
BOTH_AXES = pg.ViewBox.XYAxes
VALID_AXES_NUMBERS = (X_AXIS, Y_AXIS, BOTH_AXES)

DEFAULT_BORDER_PEN = pg.mkPen("#000000", width=1)


class ArgosPgPlotItem(PlotItem):
    """ Wrapper around pyqtgraph.graphicsItems.PlotItem

        Overrides the autoBtnClicked method.
        Middle mouse click resets the axis.

        The original PyQtGraph menu is disabled (all settings I want can be set using the config
        tree; the other settings I don't want to support). Furthermore a context menu is created
        that allows the user to rescale the axes.

        Autorange is disabled as it is expected that the (viewbox of the) plot item will be
        connected to two PgAxisRangeCti objects that control the (auto)range of the X and Y axes.

        Adds a black border of width 1.

        Sets the cursor to a cross if it's inside the viewbox.
    """

    # Use a QueuedConnection to connect to sigResetAxis so that the reset is scheduled after all
    # current events have been processed. Otherwise the mouseReleaseEvent may look for a
    # PlotCurveItem that is no longer present after the reset, which results in a RuntimeError:
    # wrapped C/C++ object of type PlotCurveItem has been deleted.
    sigResetAxis = QtSignal(int)

    def __init__(self,
                 borderPen=DEFAULT_BORDER_PEN,
                 *args, **kwargs):
        """ Constructor.

            :param borderPen: pen for drawing the viewBox border. Default black and width of 1.
        """
        super(ArgosPgPlotItem, self).__init__(*args, **kwargs)

        self.setMenuEnabled(False)
        #pg.setConfigOption('leftButtonPan', False)

        viewBox = self.getViewBox()
        viewBox.border = borderPen
        viewBox.setCursor(Qt.CrossCursor)
        viewBox.disableAutoRange(BOTH_AXES)
        viewBox.mouseClickEvent = lambda ev: self._axesMouseClickEvent(ev, BOTH_AXES)

        # Color of zoom-rectangle.
        alpha = 100
        greyVal = 160
        viewBox.rbScaleBox.setPen(pg.mkPen((0, 0, 0, alpha), width=2))
        viewBox.rbScaleBox.setBrush(pg.mkBrush(greyVal, greyVal, greyVal, alpha))

        # Add mouseClickEvent event handlers to the X and Y axis. This allows for resetting
        # the scale of each axes separately by middle mouse clicking the axis.
        for xAxisItem in (self.getAxis('bottom'), self.getAxis('top')):
            xAxisItem.mouseClickEvent = lambda ev: self._axesMouseClickEvent(ev, X_AXIS)
            xAxisItem.setCursor(Qt.SizeHorCursor)

        for yAxisItem in (self.getAxis('left'), self.getAxis('right')):
            yAxisItem.mouseClickEvent = lambda ev: self._axesMouseClickEvent(ev, Y_AXIS)
            yAxisItem.setCursor(Qt.SizeVerCursor)

        self.resetAxesAction = QtWidgets.QAction("Reset Axes", self,
                                 triggered = lambda: self.emitResetAxisSignal(BOTH_AXES),
                                 toolTip = "Resets the zoom factor of the X-axis and Y-axis")
        self.addAction(self.resetAxesAction)

        self.resetXAxisAction = QtWidgets.QAction("Reset X-axis", self,
                                 triggered = lambda: self.emitResetAxisSignal(X_AXIS),
                                 toolTip = "Resets the zoom factor of the X-axis")
        self.addAction(self.resetXAxisAction)

        self.resetYAxisAction = QtWidgets.QAction("Reset Y-axis", self,
                                 triggered = lambda: self.emitResetAxisSignal(Y_AXIS),
                                 toolTip = "Resets the zoom factor of the Y-axis")
        self.addAction(self.resetYAxisAction)



    def close(self):
        """ Is called before destruction. Can be used to clean-up resources
            Could be called 'finalize' but PlotItem already has a close so we reuse that.
        """
        logger.debug("Finalizing: {}".format(self))
        super(ArgosPgPlotItem, self).close()


    def _axesMouseClickEvent(self, mouseClickEvent, axisNumber):
        """ Handles the mouse clicks events when clicked on the axes or in the central viewbox

            :param mouseClickEvent: pyqtgraph.GraphicsScene.mouseEvents.MouseClickEvent
            :param axisNumber: the axis (X=0, Y=1, both=2)

            Emits sigResetAxis when the middle mouse button is clicked.
        """
        if mouseClickEvent.button() == QtCore.Qt.MiddleButton:
            mouseClickEvent.accept()
            self.emitResetAxisSignal(axisNumber)


    def mouseClickEvent(self, mouseClickEvent):
        """ Handles (PyQtGraph) mouse click events.

            Opens the context menu if a right mouse button was clicked. (We can't simply use
            setContextMenuPolicy(Qt.ActionsContextMenu because the PlotItem class does not derive
            from QWidget).

            :param mouseClickEvent: pyqtgraph.GraphicsScene.mouseEvents.MouseClickEvent
        """
        if mouseClickEvent.button() == QtCore.Qt.RightButton:
            contextMenu = QtWidgets.QMenu()
            for action in self.actions():
                contextMenu.addAction(action)

            screenPos = mouseClickEvent.screenPos() # Screenpos is a QPointF, convert to QPoint.
            screenX = round(screenPos.x())
            screenY = round(screenPos.y())
            contextMenu.exec_(QtCore.QPoint(screenX, screenY))


    def autoBtnClicked(self):
        """ Hides the button but does not enable/disable autorange.
            That will be done by PgAxisRangeCti
        """
        self.resetAxesAction.trigger()


    def emitResetAxisSignal(self, axisNumber):
        """ Emits the sigResetAxis with the axisNumber as parameter
            axisNumber should be 0 for X, 1 for Y, and 2 for both axes.
        """
        assert axisNumber in (VALID_AXES_NUMBERS), \
            "Axis Nr should be one of {}, got {}".format(VALID_AXES_NUMBERS, axisNumber)

        # Hide 'auto-scale (A)' button
        logger.debug("ArgosPgPlotItem.autoBtnClicked, mode:{}".format(self.autoBtn.mode))
        if self.autoBtn.mode == 'auto':
            self.autoBtn.hide()
        else:
            # Does this occur?
            msg = "Unexpected autobutton mode: {}".format(self.autoBtn.mode)
            if DEBUGGING:
                raise ValueError(msg)
            else:
                logger.warning(msg)

        logger.debug("Emitting sigResetAxis({}) for {!r}".format(axisNumber, self))
        self.sigResetAxis.emit(axisNumber)


    def setRectangleZoomOn(self, boolean):
        """ Turns on rectangle zoom mode.

            In rectangle zoom mode, the left mouse drag draws a rectangle, which defines the new
            view limits. In panning mode (the default), a left mouse drag will pan the scene.

            :param boolean: if True plotting is set to 'rect'
        """
        viewbox = self.getViewBox()
        if boolean:
            viewbox.setMouseMode(pg.ViewBox.RectMode)
        else:
            viewbox.setMouseMode(pg.ViewBox.PanMode)
